# [1] http://aip.scitation.org/doi/abs/10.1063/1.2720838

import numpy as np
from scipy.interpolate import splprep, splev

from pysisyphus.cos.ChainOfStates import ChainOfStates

class SimpleZTS(ChainOfStates):

    def __init__(self, images, param="equal", **kwargs):
        self.param = param
        super(SimpleZTS, self).__init__(images, **kwargs)

    def reparametrize(self):
        def weight_function(mean_energies):
            mean_energies = np.abs(mean_energies)
            weights = mean_energies / mean_energies.max()
            weights = np.sqrt(weights)
            return weights

        reshaped = self.coords.reshape(-1, self.coords_length)
        # To use splprep we have to transpose the coords.
        transp_coords = reshaped.transpose()

        # Equal arc length parametrization, default.
        u = None
        # Energy weighted arc length parametrization.
        if self.param == "energy":
            energies = [img.energy for img in self.images]
            mean_energies = [(energies[i] + energies[i-1])/2
                             for i in range(1, len(self.images))
            ]
            weights = weight_function(mean_energies)
            arc_segments = [0, ]
            for i in range(1, len(self.images)):
                coord_diff = np.linalg.norm(
                    self.images[i].coords - self.images[i-1].coords
                )
                next_segment = arc_segments[-1] + weights[i-1]*coord_diff
                arc_segments.append(next_segment)
            arc_segments = np.array(arc_segments)
            arc_segments /= arc_segments.max()

            u = arc_segments
        # Use chunks of 9 dimension because splprep can handle at max
        # 11 dimensions (as of scipy 1.0.0). Every atoms already yields
        # three dimensions (the coordinates X, Y, Z).
        #
        # For dim <= 11 it would go like this:
        #
        # tck, u = splprep(transp_coords, u=u, s=0)
        # uniform_mesh = np.linspace(0, 1, num=len(self.images))
        # new_points = np.array(splev(uniform_mesh, tck))
        tcks, us = zip(*[splprep(transp_coords[i:i+9], u=u, s=0)
                         for i in range(0, len(transp_coords), 9)]
        )
        uniform_mesh = np.linspace(0, 1, num=len(self.images))
        # Reparametrize mesh
        new_points = np.vstack([splev(uniform_mesh, tck) for tck in tcks])
        # Flatten along first dimension.
        new_points = new_points.reshape(-1, len(self.images))
        self.coords = new_points.transpose().flatten()

        return True



from typing import Optional

import numpy as np

from pysisyphus.Geometry import Geometry
from pysisyphus.intcoords.exceptions import NeedNewInternalsException
from pysisyphus.optimizers.closures import bfgs_multiply, get_update_mu_reg
from pysisyphus.optimizers.hessian_updates import double_damp
from pysisyphus.optimizers.Optimizer import Optimizer
from pysisyphus.optimizers.restrict_step import scale_by_max_step
from pysisyphus.optimizers.poly_fit import poly_line_search


class LBFGS(Optimizer):
    def __init__(
        self,
        geometry: Geometry,
        keep_last: int = 7,
        beta: float = 1,
        max_step: float = 0.2,
        double_damp: bool = True,
        gamma_mult: bool = False,
        line_search: bool = False,
        mu_reg: Optional[float] = None,
        max_mu_reg_adaptions: int = 10,
        control_step: bool = True,
        **kwargs,
    ) -> None:
        """Limited-memory BFGS optimizer.

        See [1] Nocedal, Wright - Numerical Optimization, 2006 for a general
        discussion of LBFGS. See pysisyphus.optimizers.hessian_updates for
        the references related to double damping and pysisyphus.optimizers.closures
        for references related to regularized LBFGS.

        Parameters
        ----------
        geometry
            Geometry to be optimized.
        keep_last
            History size. Keep last 'keep_last' steps and gradient differences.
        beta
            Force constant β in -(H + βI)⁻¹g.
        max_step
            Upper limit for the absolute component of the step vector in whatever
            unit the optimization is carried out.
        double_damp
            Use double damping procedure to modify steps s and gradient differences y
            to ensure sy > 0.
        gamma_mult
            Estimate β from previous cycle. Eq. (7.20) in [1]. See 'beta' argument.
        line_search
            Enable implicit linesearches.
        mu_reg
            Initial guess for regularization constant in regularized LBFGS.
        max_mu_reg_adaptions
            Maximum number of trial steps in regularized LBFGS.
        control_step
            Wheter to scale down the proposed step its biggest absolute component
            is equal to or below 'max_step'

        Other Parameters
        ----------------
        **kwargs
            Keyword arguments passed to the Optimizer baseclass.
        """

        self.coord_diffs = list()
        self.grad_diffs = list()
        super().__init__(geometry, max_step=max_step, **kwargs)

        self.beta = beta
        self.keep_last = int(keep_last)
        self.double_damp = double_damp
        self.gamma_mult = gamma_mult
        self.mu_reg = mu_reg
        self.max_mu_reg_adaptions = max_mu_reg_adaptions
        self.line_search = (not self.is_cos) and line_search
        self.control_step = control_step

        self.tot_adapt_mu_cycles = 0
        if self.mu_reg:
            self.mu_reg_0 = self.mu_reg  # Backup value
            self.update_mu_reg = get_update_mu_reg(logger=self.logger)
            self.control_step = False
            self.double_damp = False
            self.line_search = False
            self.log(
                f"Regularized-L-BFGS (μ_reg={self.mu_reg:.6f}) requested.\n"
                f"Disabling double damping, step control and line search."
            )

    def reset(self):
        self.coord_diffs = list()
        self.grad_diffs = list()

    def _get_opt_restart_info(self):
        opt_restart_info = {
            "coord_diffs": np.array(self.coord_diffs).tolist(),
            "grad_diffs": np.array(self.grad_diffs).tolist(),
            "double_damp": self.double_damp,
            "gamma_mult": self.gamma_mult,
            "keep_last": self.keep_last,
        }
        return opt_restart_info

    def _set_opt_restart_info(self, opt_restart_info):
        self.coord_diffs = [np.array(cd) for cd in opt_restart_info["coord_diffs"]]
        self.grad_diffs = [np.array(gd) for gd in opt_restart_info["grad_diffs"]]
        for attr in ("double_damp", "gamma_mult", "keep_last"):
            setattr(self, attr, opt_restart_info[attr])

    def get_lbfgs_step(self, forces):
        return bfgs_multiply(
            self.coord_diffs,
            self.grad_diffs,
            forces,
            beta=self.beta,
            gamma_mult=self.gamma_mult,
            mu_reg=self.mu_reg,
            logger=self.logger,
        )

    def optimize(self):
        if self.is_cos and self.align:
            rot_vecs, rot_vec_lists, _ = self.fit_rigid(
                vector_lists=(
                    self.steps,
                    self.forces,
                    self.coord_diffs,
                    self.grad_diffs,
                ),
            )
            rot_steps, rot_forces, rot_coord_diffs, rot_grad_diffs = rot_vec_lists
            self.steps = rot_steps
            self.forces = rot_forces
            self.coord_diffs = rot_coord_diffs
            self.grad_diffs = rot_grad_diffs

        forces = self.geometry.forces
        self.forces.append(forces)
        energy = self.geometry.energy
        self.energies.append(energy)
        norm = np.linalg.norm(forces)
        if not self.is_cos:
            self.log(f"      Energy={energy: >24.6f} au")
        self.log(f"norm(forces)={norm: >24.6f} au / bohr (rad)")

        if self.cur_cycle > 0 and (self.forces[-2].size == forces.size):
            y = self.forces[-2] - forces
            s = self.steps[-1]
            if self.double_damp:
                s, y = double_damp(
                    s, y, s_list=self.coord_diffs, y_list=self.grad_diffs
                )
            self.grad_diffs.append(y)
            self.coord_diffs.append(s)

            # Drop superfluous oldest vectors
            self.coord_diffs = self.coord_diffs[-self.keep_last :]
            self.grad_diffs = self.grad_diffs[-self.keep_last :]

        ###############
        # Line search #
        ###############

        ip_gradient, ip_step = None, None
        if self.line_search and (self.cur_cycle > 0):
            ip_energy, ip_gradient, ip_step = poly_line_search(
                energy, self.energies[-2], -forces, -self.forces[-2], self.steps[-1]
            )
        # Use the interpolated gradient for the step if interpolation succeeded
        if (ip_gradient is not None) and (ip_step is not None):
            forces = -ip_gradient
            self.log("Interpolation succeeded")
        # Keep the original gradient when the interpolation failed, but use
        # zero (interpolated) step.
        else:
            ip_step = np.zeros_like(forces)

        step = self.get_lbfgs_step(forces)

        # Skip adapation mu_reg in first cycle, because in this implementation
        # the first step will be steepest descent, which is independent of
        # self.mu_reg, so this loop would never break.
        adapt_mu_cycles = 0
        while self.mu_reg and (self.cur_cycle > 0):
            self.log(
                f"Adapt μ_reg={self.mu_reg:.6f}, norm(step)={np.linalg.norm(step):.6f}"
            )
            if adapt_mu_cycles == self.max_mu_reg_adaptions:
                raise Exception("Adapation of mu_reg failed! Breaking!")
            try:
                trial_energy = self.geometry.get_energy_at(self.geometry.coords + step)
                self.mu_reg, recompute_step = self.update_mu_reg(
                    self.mu_reg, energy, trial_energy, -forces, step
                )
            except NeedNewInternalsException:
                self.log("Internal coordinate breakdown in linesearch!")
                # Nothing further is done here, as the coordinates will probably also
                # breakdown when taking the step, which is then handled.
                recompute_step = False

            # Leave loop if step was accepted
            if not recompute_step:
                self.log(f"Next μ_reg={self.mu_reg:.6f}")
                break
            # Otherwise, recompute using updated μ_reg
            step = self.get_lbfgs_step(forces)
            adapt_mu_cycles += 1

        if self.mu_reg:
            self.tot_adapt_mu_cycles += adapt_mu_cycles + 1

        # Form full step. If we did not interpolate or interpolation failed,
        # ip_step will be zero.
        step = step + ip_step

        # Only try to scale down first step in regularized L-BFGS
        if (self.mu_reg and self.cur_cycle == 0) or self.control_step:
            step = scale_by_max_step(step, self.max_step)

        return step

    def postprocess_opt(self):
        if self.mu_reg:
            msg = f"\nNumber of μ updates: {self.tot_adapt_mu_cycles}"
            self.log(msg)



#!/usr/bin/env python3
"""Validate intro-template headings for key docs pages (EN/JA)."""

from __future__ import annotations

from pathlib import Path


REPO_ROOT = Path(__file__).resolve().parents[1]
DOCS_ROOT = REPO_ROOT / "docs"

TARGETS = (
    "all",
    "scan",
    "scan2d",
    "scan3d",
    "tsopt",
    "freq",
    "opt",
    "path_opt",
    "path_search",
    "irc",
    "dft",
)

REQUIRED_EN = (
    "## Minimal example",
    "## Output checklist",
    "## Common examples",
)
REQUIRED_JA = (
    "## 最小例",
    "## 出力の見方",
    "## よくある例",
)


def _heading_positions(path: Path) -> dict[str, int]:
    positions: dict[str, int] = {}
    for i, line in enumerate(path.read_text(encoding="utf-8").splitlines(), start=1):
        if line.startswith("## "):
            positions.setdefault(line.strip(), i)
    return positions


def _check_one(path: Path, required: tuple[str, ...], usage_heading: str, errors: list[str]) -> None:
    if not path.exists():
        errors.append(f"{path}: missing file")
        return

    positions = _heading_positions(path)
    missing = [h for h in required if h not in positions]
    if missing:
        errors.append(f"{path}: missing headings: {', '.join(missing)}")
        return

    req_lines = [positions[h] for h in required]
    if req_lines != sorted(req_lines):
        pairs = ", ".join(f"{h}@{positions[h]}" for h in required)
        errors.append(f"{path}: intro heading order mismatch ({pairs})")

    usage_line = positions.get(usage_heading)
    if usage_line is not None:
        late = [h for h in required if positions[h] > usage_line]
        if late:
            errors.append(
                f"{path}: intro headings must be before '{usage_heading}' "
                f"(late: {', '.join(late)})"
            )


def main() -> int:
    errors: list[str] = []

    for name in TARGETS:
        _check_one(
            DOCS_ROOT / f"{name}.md",
            REQUIRED_EN,
            "## Usage",
            errors,
        )
        _check_one(
            DOCS_ROOT / "ja" / f"{name}.md",
            REQUIRED_JA,
            "## 使用法",
            errors,
        )

    if errors:
        print("[intro-check] failed:")
        for e in errors:
            print(f"- {e}")
        return 1

    print(f"[intro-check] validated {len(TARGETS) * 2} pages.")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



import random

import numpy as np

class TableFormatter:

    def __init__(self, header, fmts, min_width=7, space=3):
        self.min_width = min_width + (space-1)
        self.space = space

        # Get lengths of header strings
        widths = np.array([len(h) for h in header])
        # Expand entries smaller than min_widht to min_width
        smaller_indices = widths < min_width
        widths[smaller_indices] = min_width
        self.widths = widths

        # Construct header
        #header_fmts = ["{:" + "{}".format(width) + "s}"
        #               for width in self.widths]
        header_fmts = self.min_width_fmts()
        self._header = self.join_format(header_fmts, header)
        self._header += "\n" + (self.space*" ").join(
                ["-"*width for width in self.widths]
        )

        # Modify fmts to consinder min_widths 
        self.fmts = self.min_width_fmts(fmts)

    def min_width_fmts(self, raw_fmts=None):
        if not raw_fmts:
            raw_fmts = ["s" for _ in self.widths]
        return ["{:>" + "{}".format(width) + fmt + "}"
                for width, fmt in zip(self.widths, raw_fmts)]

    def join_format(self, fmts, lst):
        """Format a given iterable lst with formats given in the iterable
        lst and return the joined items of the formatted list."""
        return (self.space*" ").join(
            [fmt.format(item) for fmt, item in zip(fmts, lst)]
        )

    @property
    def header(self):
        return self._header

    def line(self, *args):
        #formatted = " "["
        return self.join_format(self.fmts, args)

def run():
    header = "# |dx| |tangent|".split()
    fmts = ["d", ".2E", ".3E"]
    min_width = 10
    tp = TableFormatter(header, fmts, min_width)
    print(tp.header)
    for i in range(10):
        dx = random.random()
        tangent = random.random()
        print(tp.line(i, dx, tangent))

if __name__ == "__main__":
    run()



import warnings

import numpy as np

try:
    from openbabel import openbabel as ob
    from openbabel import pybel
except ModuleNotFoundError:
    pass
except ImportError:
    pass

from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.constants import BOHR2ANG, AU2KJPERMOL, AU2KCALPERMOL


def _getpluginnames(ptype):
    plugins = ob.vectorString()
    ob.OBPlugin.ListAsVector(ptype, None, plugins)
    return [p.split()[0].lower() for p in plugins if p.strip()]


def _getplugins(findplugin, names):
    return dict([(x, findplugin(x)) for x in names if findplugin(x)])


class OBabel(Calculator):
    conv_dict = {
        "kj/mol": AU2KJPERMOL,
        "kcal/mol": AU2KCALPERMOL,
    }

    def __init__(self, ff="gaff", mol=None, **kwargs):
        super().__init__(**kwargs)

        if self.charge != 0:
            warnings.warn("'charge' keyword is currently ignored!")

        self.ff_name = ff
        avail_ffs = _getplugins(
            ob.OBForceField.FindType, _getpluginnames("forcefields")
        )
        self.ff = avail_ffs[self.ff_name]
        self.mol = mol

        unit = self.ff.GetUnit()
        an_grad = self.ff.HasAnalyticalGradients()
        self.log(f"Energy unit: {unit}, analytical gradient: {an_grad}")
        # Used to convert energy from force field units to atomic units
        self.conv_fac = self.conv_dict[unit.lower()]

        self.is_setup = False

    def setup(self, atoms, coords):
        xyz = self.prepare_xyz_string(atoms, coords)
        if self.mol is None:
            self.mol = pybel.readstring("xyz", xyz)
        else:
            self.mol.OBMol.SetCoordinates(ob.double_array(coords*BOHR2ANG))

        if not self.is_setup:
            self.is_setup = self.ff.Setup(self.mol.OBMol)
            assert self.is_setup
        else:
            self.ff.SetCoordinates(self.mol.OBMol)
        return self.mol

    def get_energy(self, atoms, coords, **prepare_kwargs):
        self.setup(atoms, coords)
        energy = self.ff.Energy() / self.conv_fac

        results = {"energy": energy}
        return results

    def get_forces(self, atoms, coords, **prepare_kwargs):
        mol = self.setup(atoms, coords)
        ff = self.ff

        energy = ff.Energy(True) / self.conv_fac
        grad = [ff.GetGradient(atom.OBAtom) for atom in mol.atoms]
        grad = np.array([(g.GetX(), g.GetY(), g.GetZ()) for g in grad])
        # WTH?`It seems openbabel does not return the gradient, but the force...
        # So we don't multiply with -1 here...
        forces = grad.flatten() * BOHR2ANG / self.conv_fac

        results = {"energy": energy, "forces": forces}
        return results



import numpy as np

from pysisyphus.calculators.Calculator import Calculator


class LennardJones(Calculator):

    # Corresponds to σ = 1 Å, as the default value in ASE, but
    # pysisyphus uses au/Bohr.
    def __init__(self, sigma=1.8897261251, epsilon=1, rc=None):
        super().__init__()

        self.sigma = sigma
        self.epsilon = epsilon
        # Cutoff distance
        if rc is None:
            rc = 3 * self.sigma
        self.rc = rc
        # Shift energy
        self.e0 = (4 * self.epsilon *
                   ((self.sigma/self.rc)**12 - (self.sigma/self.rc)**6)
        )

    def calculate(self, coords3d):
        # Index pairs
        a, b = np.triu_indices(len(coords3d), 1)

        # Distances
        diffs = coords3d[a] - coords3d[b]  # Shape: (N_atoms, 3)
        rs = np.linalg.norm(diffs, axis=1)
        c6 = np.where(rs <= self.rc, (self.sigma/rs)**6, np.zeros_like(rs))
        energy = -self.e0 * (c6 != 0.0).sum()
        c12 = c6**2
        energy += np.sum(4*self.epsilon * (c12 - c6))

        """
        Gradient related remarks:

        Lennard-Jones-potential:
            LJ(r) =
                4*ε*[(σ/r)**12 - (σ/r)**6]

        Derivative of quotient appearing in LJ potential w.r.t first cartesian
        coordinate:
            d(σ/r)**n/dx_1
                = σ**n * d/dx_1 r**(-n)
                = σ**n * (-n/2) * (2x1-2x2) * r**(-n) * r**(-2)
                = σ**n * r**(-n) * (-n) * (x1-x2) * r**(-2)
                = (σ/r)**n * (-n) * (x1-x2) / r**2

        Derivate w.r.t to cartesian x coordinate of atom A (x_1):
            dLJ(r)/dx_1 = 
                24*ε*[-2*(σ/r)**12 + (σ/r)**6]*(x1-x2)/r**2

        Derivate w.r.t to cartesian x coordinate of atom B (x_2):
            dLJ(r)/dx_2 = 
                -24*ε*[-2*(σ/r)**12 + (σ/r)**6]*(x1-x2)/r**2

        The derivate w.r.t to x_2 differs only by a factor of -1!
        """
        prefactors = 24*self.epsilon * (c6 - 2*c12) / rs**2
        products = prefactors[:,None] * diffs

        gradient = np.zeros_like(coords3d)
        # Every pair (a, b) contributes to the total gradient of atoms a and b.
        for i, prod in enumerate(products):
            gradient[a[i]] += prod
            gradient[b[i]] -= prod

        return energy, -gradient

    def get_energy(self, atoms, coords):
        energy, _ = self.calculate(coords.reshape(-1, 3))
        return {"energy": energy}

    def get_forces(self, atoms, coords):
        energy, forces = self.calculate(coords.reshape(-1, 3))
        return {"energy": energy,
                "forces": forces.flatten(),
        }



# See [1] https://pubs.acs.org/doi/pdf/10.1021/j100247a015
#         Banerjee, 1985
#     [2] https://aip.scitation.org/doi/abs/10.1063/1.2104507
#         Heyden, 2005
#     [3] https://onlinelibrary.wiley.com/doi/abs/10.1002/jcc.540070402
#         Baker, 1985
#     [4] https://link.springer.com/article/10.1007/s002140050387
#         Besalu, 1998


import numpy as np

from pysisyphus.tsoptimizers.TSHessianOptimizer import TSHessianOptimizer


class RSPRFOptimizer(TSHessianOptimizer):
    def optimize(self):
        energy, gradient, H, eigvals, eigvecs, resetted = self.housekeeping()
        self.update_ts_mode(eigvals, eigvecs)

        # Transform gradient to eigensystem of hessian
        gradient_trans = eigvecs.T.dot(gradient)
        # Minimize energy along all modes, except the TS-mode
        min_indices = [i for i in range(gradient_trans.size) if i not in self.roots]
        # Maximize energy along all requested modes.
        max_indices = [i for i in range(gradient_trans.size) if i in self.roots]
        # Get line search steps, if requested.
        ip_step_trans, gradient_trans = self.step_and_grad_from_line_search(
            energy,
            gradient_trans,
            eigvecs,
            min_indices,
            max_indices,
        )

        """In the RS-(P)RFO method we have to scale the matrices with alpha.
        Unscaled matrix (Eq. 8) in [1]:
            (H  g) (x)          (S 0) (x)
                       = lambda
            (g+ 0) (1)          (0 1) (1)
        with
            S = alpha * Identity matrix
        and multiplying from the left with the inverse of the scaling matrix
            (1/alpha 0)

            (0       1)
        we get
            (1/alpha 0) (H  g) (x)          (x)
                                   = lambda
            (0       1) (g+ 0) (1)          (1)
        eventually leading to the scaled matrix:
            (H/alpha  g/alpha) (x)          (x)
                                   = lambda     .
            (g+             0) (1)          (1)
        """

        alpha = self.alpha0
        for mu in range(self.max_micro_cycles):
            self.log(f"RS-PRFO micro cycle {mu:02d}, alpha={alpha:.6f}")

            # Maximize energy along the chosen TS mode. The matrix is hardcoded
            # as 2x2, so only first-order saddle point searches are supported.
            # Try secular equation solver first (O(N) vs O(N^3))
            secular_max = self.solve_rfo_secular(
                eigvals[max_indices], gradient_trans[max_indices], alpha,
                kind="max", prev_eigvec=self.prev_eigvec_max,
            )
            if secular_max is not None:
                step_max, eigval_max, nu_max, self.prev_eigvec_max = secular_max
            else:
                H_aug_max = self.get_augmented_hessian(
                    eigvals[max_indices], gradient_trans[max_indices], alpha
                )
                step_max, eigval_max, nu_max, self.prev_eigvec_max = self.solve_rfo(
                    H_aug_max, "max", prev_eigvec=self.prev_eigvec_max
                )

            # Minimize energy along all modes, but the TS mode.
            secular_min = self.solve_rfo_secular(
                eigvals[min_indices], gradient_trans[min_indices], alpha,
                kind="min", prev_eigvec=self.prev_eigvec_min,
            )
            if secular_min is not None:
                step_min, eigval_min, nu_min, self.prev_eigvec_min = secular_min
            else:
                H_aug_min = self.get_augmented_hessian(
                    eigvals[min_indices], gradient_trans[min_indices], alpha
                )
                step_min, eigval_min, nu_min, self.prev_eigvec_min = self.solve_rfo(
                    H_aug_min, "min", prev_eigvec=self.prev_eigvec_min
                )

            # Calculate overlap between directions over the course of the micro cycles
            # if mu == 0:
            # TODO: convert back to original space
            # ref_step_max = step_max.copy()
            # ref_step_min = step_min.copy()
            min_norm = np.linalg.norm(step_min)
            max_norm = np.linalg.norm(step_max)
            self.log(f"norm(step_max)={max_norm:.6f}")
            self.log(f"norm(step_min)={min_norm:.6f}")
            self.log(f"norm(step_max)/norm(step_min)={max_norm/min_norm:.2%}")
            # Calculate overlaps with originally proposed step in mu == 0
            # TODO: convert back to original space
            # max_ovlp = ref_step_max @ step_max
            # min_ovlp = ref_step_min @ step_min

            # As of Eq. (8a) of [4] max_eigval and min_eigval also
            # correspond to:
            # max_eigval = -forces_trans[max_indices].dot(max_step)
            # min_eigval = -forces_trans[min_indices].dot(min_step)

            # Create the full PRFO step
            step = np.zeros_like(gradient_trans)
            step[max_indices] = step_max
            step[min_indices] = step_min
            step_norm = np.linalg.norm(step)
            self.log(f"norm(step)={step_norm:.6f}")

            inside_trust = step_norm <= self.trust_radius
            if inside_trust:
                self.log(
                    "Restricted step satisfies trust radius of "
                    f"{self.trust_radius:.6f}"
                )
                self.log(
                    f"Micro-cycles converged in cycle {mu:02d} with "
                    f"alpha={alpha:.6f}!"
                )
                break

            # Derivative of the squared step w.r.t. alpha
            # max subspace
            dstep2_dalpha_max = (
                2
                * eigval_max
                / (1 + step_max.dot(step_max) ** 2 * alpha)
                * np.sum(
                    gradient_trans[max_indices] ** 2
                    / (eigvals[max_indices] - eigval_max * alpha) ** 3
                )
            )
            # min subspace
            dstep2_dalpha_min = (
                2
                * eigval_min
                / (1 + step_min.dot(step_min) * alpha)
                * np.sum(
                    gradient_trans[min_indices] ** 2
                    / (eigvals[min_indices] - eigval_min * alpha) ** 3
                )
            )
            dstep2_dalpha = dstep2_dalpha_max + dstep2_dalpha_min
            # Update alpha
            alpha_step = (
                2 * (self.trust_radius * step_norm - step_norm**2) / dstep2_dalpha
            )
            alpha += alpha_step

        # Right now the step is still given in the Hessians eigensystem. We
        # transform it back now.
        step += ip_step_trans
        step = eigvecs.dot(step)
        step_norm = np.linalg.norm(step)

        # With max_micro_cycles = 1 the RS part is disabled and the step
        # probably isn't scaled correctly in the one micro cycle.
        # In this case we use a naive scaling if the step is too big.
        if (self.max_micro_cycles == 1) and (step_norm > self.trust_radius):
            step = step / step_norm * self.trust_radius
        self.log(f"norm(step)={np.linalg.norm(step):.6f}")

        # Eq. (6) from [4] seems erronous ... the prediction is usually only ~50%
        # of the actual change ...
        # predicted_energy_change = 1/2 * (eigval_max / nu_max**2 + eigval_min / nu_min**2)
        # self.predicted_energy_changes.append(predicted_energy_change)

        self.predicted_energy_changes.append(self.rfo_model(gradient, self.cur_H, step))

        self.log("")
        step = self.full_from_active(step)
        return step



# [1  ] https://aip.scitation.org/doi/pdf/10.1063/1.3514202?class=pdf
#       Original EulerPC
#       Hratchian, Schlegel, 2010
# [2  ] https://aip.scitation.org/doi/pdf/10.1063/1.1724823?class=pdf
#       Original HPC
#       Hratchian, Schlegel, 2004
# [3  ] https://pubs.rsc.org/en/content/articlepdf/2017/cp/c7cp03722h
#       EulerPC re-implementation
#       Meisner, Kästner, 2017
# [3.1] http://www.rsc.org/suppdata/c7/cp/c7cp03722h/c7cp03722h1.pdf
#       Corresponding SI
# [4  ] https://aip.scitation.org/doi/pdf/10.1063/1.3593456?class=pdf
#       Hratchian, Frisch, 2011
# [6  ] https://aip.scitation.org/doi/10.1063/1.3593456<Paste>
#       Hratchian, Frisch
# 	Further improvements for DWI; not implemented

import time
from collections import deque

import numpy as np

from pysisyphus.helpers_pure import rms
from pysisyphus.irc.DWI import DWI
from pysisyphus.irc.IRC import IRC
from pysisyphus.optimizers.hessian_updates import bfgs_update, bofill_update

import torch


class EulerPC(IRC):
    def __init__(
        self,
        *args,
        hessian_recalc=None,
        hessian_update="bofill",
        hessian_init="calc",
        max_pred_steps=500,
        loose_cycles=3,
        corr_func="mbs",
        **kwargs,
    ):
        super().__init__(*args, hessian_init=hessian_init, **kwargs)

        self.hessian_recalc = hessian_recalc
        self.hessian_update = {
            "bfgs": bfgs_update,
            "bofill": bofill_update,
        }
        self.hessian_update_func = self.hessian_update[hessian_update]
        self.max_pred_steps = int(max_pred_steps)
        self.loose_cycles = loose_cycles

        corr_funcs = {
            "mbs": self.corrector_step,
        }
        self.corr_func = corr_funcs[corr_func]

        self._to_active_vec = lambda v: v[self._act_dofs]

        def _to_full_vec(v_act, template):
            if isinstance(v_act, torch.Tensor):
                full = template.clone()
            else:
                full = template.copy()
            full[self._act_dofs] = v_act
            return full

        self._to_full_vec = _to_full_vec
        
    def prepare(self, *args, **kwargs):
        super().prepare(*args, **kwargs)

        # Initialize the distance weighted interpolator with the data
        # from the initial displacement.
        self.dwi = DWI()
        mw_grad_full = self.mw_gradient
        energy = self.energy

        mw_grad_act   = self._to_active_vec(mw_grad_full)
        mw_coords_act = self._to_active_vec(self.mw_coords)

        if isinstance(self.mw_hessian, torch.Tensor):
            self.dwi.update(mw_coords_act, energy, mw_grad_act, self.mw_hessian.detach().clone())
        else:
            self.dwi.update(mw_coords_act, energy, mw_grad_act, self.mw_hessian.copy())

        if self.downhill:
            return

        # Do a first Hessian update with the information between the TS
        # and the initially displaced geometry.
        dx = self._to_active_vec(self.mw_coords - self.ts_mw_coords)
        dg = mw_grad_act - self._to_active_vec(self.ts_mw_gradient)
        dH, key = self.hessian_update_func(self.mw_hessian, dx, dg)
        self.log(f"Did {key} hessian update.")
        self.mw_hessian += dH

    def get_integration_length_func(self, init_mw_coords):
        """
        Return a closure that computes the un-mass-weighted path length Δs.
        """
        if init_mw_coords.shape[0] == self._m_sqrt.shape[0]:
            m_sqrt_vec = self._m_sqrt                    # full (3N)
        else:
            m_sqrt_vec = self._m_sqrt[self._act_dofs]    # active-only

        def get_integration_length(cur_mw_coords):
            return np.linalg.norm((cur_mw_coords - init_mw_coords) / m_sqrt_vec)

        return get_integration_length

    def step(self):
        ##################
        # PREDICTOR STEP #
        ##################

        mw_grad_full = self.mw_gradient            # 3N-vector, mass-weighted
        energy       = self.energy
        mw_grad_act  = self._to_active_vec(mw_grad_full)  # active slice

        if self.cur_cycle > 0:
            if self.hessian_recalc and (self.cur_cycle % self.hessian_recalc == 0):
                H_new = self.geometry.hessian
                act_n = len(self._act_dofs)
                if H_new.shape[0] != act_n:
                    H_new = H_new[self._act_dofs][:, self._act_dofs]
                self.mw_hessian = self._mw_hessian_active(H_new)
                self.geometry.clear()
                self.log("Calculated exact hessian")
            else:
                dx = self._to_active_vec(self.mw_coords - self.irc_mw_coords[-2])
                dg = mw_grad_act - self._to_active_vec(self.irc_mw_gradients[-2])
                dH, key = self.hessian_update_func(self.mw_hessian, dx, dg)
                if isinstance(dH, torch.Tensor):
                    self.mw_hessian.add_(dH)
                else:
                    self.mw_hessian += dH
                self.log(f"Did {key} hessian update before predictor step.")
            if isinstance(self.mw_hessian, torch.Tensor):
                self.dwi.update(
                    self._to_active_vec(self.mw_coords).copy(),
                    energy, mw_grad_act,
                    self.mw_hessian.detach().clone()
                )
            else:
                self.dwi.update(
                    self._to_active_vec(self.mw_coords).copy(),
                    energy, mw_grad_act,
                    self.mw_hessian.copy()
                )

        # Create a copy of the inital coordinates for the determination
        # of the actual step size in the predictor Euler integration.
        init_mw_coords = self.mw_coords.copy()

        get_integration_length = self.get_integration_length_func(init_mw_coords)

        # Calculate predictor Euler-integration step length. See get_conv_fact
        # method definition for a comment on this.
        conv_fact = self.get_conv_fact(mw_grad_act)
        euler_step_length = self.step_length / (self.max_pred_steps / conv_fact)

        def taylor_gradient(step_full):
            """Return full-length gradient via 2nd-order Taylor (active Hessian)."""
            step_act = self._to_active_vec(step_full)
            if isinstance(self.mw_hessian, torch.Tensor):
                step_t  = torch.tensor(step_act, dtype=self.mw_hessian.dtype,
                                       device=self.mw_hessian.device)
                hvp_act = (self.mw_hessian @ step_t).cpu().numpy()
            else:
                hvp_act = self.mw_hessian @ step_act
            grad_act = mw_grad_act + hvp_act
            return self._to_full_vec(grad_act, step_full)

        # These variables will hold the coordinates and gradients along
        # the Euler integration and will be updated frequently.
        euler_mw_coords = self.mw_coords.copy()
        euler_mw_grad   = mw_grad_full.copy()
        self.log(
            f"Predictor-Euler-integration with Δs={euler_step_length:.6f} "
            f"for up to {self.max_pred_steps} steps\n     #  |step|  d|step|"
        )
        prev_cur_length = 0.0
        for i in range(self.max_pred_steps):
            # Calculate step length in non-mass-weighted coordinates
            cur_length = get_integration_length(euler_mw_coords)
            if i % 50 == 0:
                diff = cur_length - prev_cur_length
                self.log(f"\t{i:03d}: {cur_length:.4f} Δ={diff:.4f}")
                prev_cur_length = cur_length

            # Check if we achieved the desired step length.
            if cur_length >= self.step_length:
                self.log(
                    "Predictor-Euler integration converged with "
                    f"Δs={cur_length:.4f} (desired Δs={self.step_length:.4f}) "
                    f"after {i+1} steps!"
                )
                break
            grad_norm = np.linalg.norm(euler_mw_grad)
            if not np.isfinite(grad_norm) or grad_norm == 0.0:
                self.log("Gradient norm is zero/NaN; using transition vector direction.")
                direction = getattr(self, "mw_transition_vector", euler_mw_grad)
                dir_norm = np.linalg.norm(direction)
                if not np.isfinite(dir_norm) or dir_norm == 0.0:
                    raise ValueError("Cannot determine IRC step direction (zero/NaN norms).")
                step_ = euler_step_length * -direction / dir_norm
            else:
                step_ = euler_step_length * -euler_mw_grad / grad_norm
            euler_mw_coords += step_
            # Determine actual step by comparing the current and the initial coordinates
            euler_step = euler_mw_coords - init_mw_coords
            euler_mw_grad = taylor_gradient(euler_step)
        else:
            self.log(
                f"Predictor-Euler integration did not converge in {i+1} "
                f"steps. Δs={cur_length:.4f}."
            )

            # Check if we are already sufficiently converged. If so signal
            # convergence.
            self.mw_coords = euler_mw_coords

            # Use rms of gradient from taylor expansion for convergence check.
            euler_grad = self.unweight_vec(euler_mw_grad)
            rms_grad = rms(euler_grad)

            # Or check true gradient? But this would need an additional calculation,
            # so I disabled it for now.
            # rms_grad = rms(self.gradient)

            if self.cur_cycle < self.loose_cycles:
                self.log(
                    f"Current cycle {self.cur_cycle} is still in 'loose' mode.\n"
                    "Continuing IRC integration even though predictor integration "
                    f"did not succeed.\n{self.loose_cycles - self.cur_cycle - 1} loose "
                    "cycles remaining."
                )
            # elif rms_grad <= 5*self.rms_grad_thresh:
            elif rms_grad <= self.rms_grad_thresh:
                self.log("Sufficient convergence achieved on rms(grad)")
                self.converged = True
                return
        self.log("")

        # Calculate energy and gradient at new predicted geometry. Update the
        # hessian accordingly. These results will be added to the DWI for use
        # in the corrector step.
        self.mw_coords = euler_mw_coords
        self.log("Calculating energy and gradient at predictor step geometry.")
        mw_grad_full = self.mw_gradient
        energy       = self.energy
        mw_grad_act  = self._to_active_vec(mw_grad_full)

        # Hessian update
        dx = self._to_active_vec(self.mw_coords - self.irc_mw_coords[-1])
        dg = mw_grad_act - self._to_active_vec(self.irc_mw_gradients[-1])
        dH, key = self.hessian_update_func(self.mw_hessian, dx, dg)
        self.mw_hessian += dH
        self.log(f"Did {key} hessian update after predictor step.\n")
        if isinstance(self.mw_hessian, torch.Tensor):
            self.dwi.update(
                self._to_active_vec(self.mw_coords).copy(),
                energy, mw_grad_act,
                self.mw_hessian.detach().clone()
            )
        else:
            self.dwi.update(
                self._to_active_vec(self.mw_coords).copy(),
                energy, mw_grad_act,
                self.mw_hessian.copy()
            )

        corrected_mw_coords_act = self.corr_func(self._to_active_vec(init_mw_coords), self.step_length, self.dwi)
        self.mw_coords = self._to_full_vec(corrected_mw_coords_act, self.mw_coords)
        corr_step_length = get_integration_length(self.mw_coords)
        self.log(f"Corrected unweighted step length: {corr_step_length:.6f}")

    def corrector_step(self, init_mw_coords, step_length, dwi):
        self.log("Corrector step using mBS integration (active dofs)")

        get_integration_length = self.get_integration_length_func(init_mw_coords)

        errors = list()
        richardson = dict()
        for k in range(15):
            points = 20 * (2**k)
            corr_step_length = step_length / (points - 1)
            cur_coords = init_mw_coords.copy()
            # Only keep the last 2 coords (needed for oscillation check)
            k_coords = deque(maxlen=2)
            cur_length = 0

            # Integrate until the desired spacing is reached
            while True:
                k_coords.append(cur_coords.copy())
                if abs(step_length - cur_length) < 0.5 * corr_step_length:
                    self.log(
                        f"\tk={k:02d} points={points: >4d} "
                        f"step_length={corr_step_length:.4f} Δs={cur_length:.4f}"
                    )
                    break

                energy, gradient = dwi.interpolate(cur_coords, gradient=True)
                cur_coords += corr_step_length * -gradient / np.linalg.norm(gradient)
                # cur_length += corr_step_length
                cur_length = get_integration_length(cur_coords)

                # Check for oscillation
                if len(k_coords) >= 2:
                    prev_coords = k_coords[-2]
                    osc_norm = np.linalg.norm(cur_coords - prev_coords)
                    # TODO: Handle this by restarting everything with a smaller stepsize?
                    # Check 10.1039/c7cp03722h SI
                    if osc_norm <= corr_step_length:
                        self.log(
                            "\tDetected oscillation in Corrector-Euler "
                            f"integration for k={k:02d} and {points} points.\n"
                            "\tAborting corrector integration!"
                        )
                        return prev_coords
            richardson[(k, 0)] = cur_coords

            # Refine using Richardson extrapolation
            # Set additional values using Richard extrapolation
            for j in range(1, k + 1):
                richardson[(k, j)] = (
                    (2**j) * richardson[(k, j - 1)] - richardson[(k - 1, j - 1)]
                ) / (2**j - 1)
            # Can only be done after the second successful integration
            if k > 0:
                # Error estimate according to Numerical Recipes Eq. (17.3.9).
                # We compare the last two entries/columns in the current row.
                # RMS error
                error = np.sqrt(
                    np.mean((richardson[(k, k)] - richardson[(k, k - 1)]) ** 2)
                )
                errors.append(error)
                if error <= 1e-5:
                    self.log(f"mBS integration converged (error={error:.4e})!")
                    break
        else:
            raise Exception("Richardson did not converge!")

        self.log(
            f"Returning corrected mass-weighted coordinates from richardson[({k},{k})]"
        )
        return richardson[(k, k)]



#!/usr/bin/env python3
"""Verify that options forwarded from `pdb2reaction all` to `pdb2reaction scan` exist."""

from __future__ import annotations

import ast
from pathlib import Path
from typing import Set


REPO_ROOT = Path(__file__).resolve().parents[1]
ALL_PY = REPO_ROOT / "pdb2reaction" / "all.py"
SCAN_PY = REPO_ROOT / "pdb2reaction" / "scan.py"
SCAN_COMMON_PY = REPO_ROOT / "pdb2reaction" / "scan_common.py"


def _is_click_option_call(node: ast.Call) -> bool:
    func = node.func
    return (
        isinstance(func, ast.Attribute)
        and func.attr == "option"
        and isinstance(func.value, ast.Name)
        and func.value.id == "click"
    )


def _const_flag(node: ast.AST) -> str | None:
    if isinstance(node, ast.Constant) and isinstance(node.value, str):
        value = node.value.strip()
        if value.startswith("--"):
            return value
    return None


def _expand_flag_spec(spec: str) -> Set[str]:
    parts = [p.strip() for p in spec.split("/") if p.strip()]
    flags = {p for p in parts if p.startswith("--")}
    if flags:
        return flags
    return {spec}


def _collect_scan_declared_flags(scan_tree: ast.AST) -> Set[str]:
    flags: Set[str] = set()
    for node in ast.walk(scan_tree):
        if not isinstance(node, ast.Call) or not _is_click_option_call(node):
            continue
        for arg in node.args:
            flag = _const_flag(arg)
            if flag is not None:
                flags.update(_expand_flag_spec(flag))
    return flags


class _ForwardedScanFlags(ast.NodeVisitor):
    def __init__(self) -> None:
        self.in_cli = False
        self.flags: Set[str] = set()

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        was_in_cli = self.in_cli
        if node.name == "cli":
            self.in_cli = True
        self.generic_visit(node)
        self.in_cli = was_in_cli

    def visit_Assign(self, node: ast.Assign) -> None:
        if self.in_cli:
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == "scan_args":
                    if isinstance(node.value, ast.List):
                        for elt in node.value.elts:
                            flag = _const_flag(elt)
                            if flag is not None:
                                self.flags.add(flag)
        self.generic_visit(node)

    def visit_Call(self, node: ast.Call) -> None:
        if self.in_cli:
            if isinstance(node.func, ast.Name) and node.func.id == "_append_cli_arg":
                if len(node.args) >= 2 and isinstance(node.args[0], ast.Name) and node.args[0].id == "scan_args":
                    flag = _const_flag(node.args[1])
                    if flag is not None:
                        self.flags.add(flag)

            if isinstance(node.func, ast.Attribute) and node.func.attr == "append":
                if isinstance(node.func.value, ast.Name) and node.func.value.id == "scan_args" and node.args:
                    value = node.args[0]
                    flag = _const_flag(value)
                    if flag is not None:
                        self.flags.add(flag)
                    elif isinstance(value, ast.IfExp):
                        body_flag = _const_flag(value.body)
                        else_flag = _const_flag(value.orelse)
                        if body_flag is not None:
                            self.flags.add(body_flag)
                        if else_flag is not None:
                            self.flags.add(else_flag)
        self.generic_visit(node)


def main() -> int:
    all_tree = ast.parse(ALL_PY.read_text(encoding="utf-8"), filename=str(ALL_PY))
    scan_tree = ast.parse(SCAN_PY.read_text(encoding="utf-8"), filename=str(SCAN_PY))
    scan_common_tree = ast.parse(
        SCAN_COMMON_PY.read_text(encoding="utf-8"),
        filename=str(SCAN_COMMON_PY),
    )

    declared = _collect_scan_declared_flags(scan_tree)
    # `scan` receives many flags through @add_scan_common_options(...).
    declared.update(_collect_scan_declared_flags(scan_common_tree))
    collector = _ForwardedScanFlags()
    collector.visit(all_tree)
    forwarded = collector.flags

    missing = sorted(flag for flag in forwarded if flag not in declared)
    if missing:
        print("Detected all->scan option contract drift.")
        for flag in missing:
            print(f"  missing in scan CLI: {flag}")
        return 1

    print(f"[scan-contract] all forwards {len(forwarded)} scan options; all are declared.")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



# [1] https://pubs.acs.org/doi/10.1021/j100338a027
#     Koeski, Gordon, 1989
# [2] https://aip.scitation.org/doi/abs/10.1063/1.459634
#     Page, Doubleday, McIver, 1990

from collections import namedtuple

import h5py
import numpy as np
from scipy.optimize import bisect

from pysisyphus.constants import AU2KJPERMOL


def get_curv_vec(H, Gv, v0, w0):
    v0 = v0[:, None]
    I = np.eye(v0.size)
    first = np.linalg.pinv(2 * w0 * I - H, rcond=1e-8)
    second = Gv.dot(v0) - v0.T.dot(Gv).dot(v0) * v0
    v1 = first.dot(second)
    return v1.flatten()


def taylor_closure(H, Gv, v0, v1, w0):
    """Taylor expansion of energy to 3rd order.

    dx(ds) = ds*v0 + ds**2*v1 / 2
    dE(ds) = dx^T H dx / 2 + dx^T [Gv] dx / 6

    H = Hessian
    Gv = 3rd derivative of energy along v0
    v0 = Transition vector
    v1 = Curvature vector
    w0 = Eigenvalue belonging to v0
    """

    # Precontract some values that will be reused
    v0v1 = v0.dot(v1)
    v1Hv1 = v1.dot(H).dot(v1)
    v0Gvv0 = v0.dot(Gv).dot(v0)
    v0Gvv1 = v0.dot(Gv).dot(v1)
    v1Gvv1 = v1.dot(Gv).dot(v1)

    def dE(ds):
        ds2 = ds ** 2
        ds3 = ds2 * ds
        ds4 = ds2 * ds2

        # Δx^T H Δx / 2
        quad = (w0 * ds2 + ds3 * w0 ** 2 * v0v1 + (ds4 * v1Hv1) / 4) / 2
        # Δx^T [Gv] Δx / 6
        cubic = (ds2 * v0Gvv0 + ds3 * v0Gvv1 + ds4 * v1Gvv1 / 4) / 6
        return quad + cubic

    return dE


ThirdDerivResult = namedtuple(
    "ThirdDerivResult",
    "coords_plus energy_plus H_plus coords_minus energy_minus H_minus G_vec vec ds",
)


def third_deriv_fd(geom, vec, ds=0.001):
    """Third derivative of the energy in direction 'vec'."""

    def get_H(geom, coords):
        results = geom.get_energy_and_cart_hessian_at(coords)
        energy = results["energy"]
        H = results["hessian"]
        H_mw = geom.mass_weigh_hessian(H)
        # Only project for multi-atomic geometries.
        if geom.coords.size > 3:
            H_mw = geom.eckart_projection(H_mw)
        return H_mw, H, energy

    delta = ds * vec
    plus = geom.coords + delta
    minus = geom.coords - delta
    H_mw_plus, H_plus, energy_plus = get_H(geom, plus)
    H_mw_minus, H_minus, energy_minus = get_H(geom, minus)
    G_vec = (H_mw_plus - H_mw_minus) / (2 * ds)

    third_deriv_res = ThirdDerivResult(
        coords_plus=plus,
        energy_plus=energy_plus,
        H_plus=H_plus,
        coords_minus=minus,
        energy_minus=energy_minus,
        H_minus=H_minus,
        G_vec=G_vec,
        vec=vec,
        ds=ds,
    )
    return G_vec, third_deriv_res


def cubic_displ(H, v0, w0, Gv, dE):
    """
    According to Eq. (26) in [2] v1 does not depend on the sign of v0.
        v1 = (F0 - 2v0^T F0 v0 I)⁻¹ x ([G0v0] - v0^T [G0v0] v0 I) v0
    The first term is obviously independent of v0's sign. Using v0' = -v0 the
    second term becomes
        ([G0v0'] - v0'^T [G0v0'] v0' I) v0'
        (-[G0v0] - v0^T [G0v0'] v0 I) v0'
        (-[G0v0] + v0^T [G0v0] v0 I) v0'
        -(-[G0v0] + v0^T [G0v0] v0 I) v0
        ([G0v0] - v0^T [G0v0] v0 I) v0
    Strictly speaking the contraction [G0v0] should depend on the sign of v0.
    In the current implementation, the direction of v0 is not taken into account,
    but
        get_curv_vec(H, Gv, v0, w0) == get_curv_vec(H, -Gv, -v0, w0) .
    But somehow the Taylor expansion gives bogus results when called with -Gv and -v0...
    """

    assert dE < 0.0, f"Supplied dE={dE:.6f} is positive but it must be negative!"
    assert w0 < 0.0, f"Expected first eigenvalue to be negative but it is w0={w0:.6e}!"

    v1 = get_curv_vec(H, Gv, v0, w0)
    E_taylor = taylor_closure(H, Gv, v0, v1, w0)

    def func(ds):
        return E_taylor(ds) - dE

    def prepare_bisect(x0, theta=1.25, max_cycles=20):
        """Determine (lower) upper bound for scipy.optimize.bisect."""
        assert theta > 1.0

        ds = x0
        dE_min = func(0.0)
        dE_prev = dE_min
        ds_min = ds
        # Grow until we find an upper (lower) bound of the interval
        for _ in range(max_cycles):
            dE = func(ds)
            # print(
                # f"ds={ds:.4f} dE={dE*AU2KJPERMOL:.3f} dE_min={dE_min*AU2KJPERMOL:.3f}"
            # )

            if dE <= 0.0:
                break
            # Keep best guess
            elif dE <= dE_min:
                dE_min = dE
                ds_min = ds
            # Return (yet) best guess when function value grows again
            elif dE >= dE_prev:
                ds = ds_min
                break

            dE_prev = dE
            ds *= theta  # Grow ds
        return ds

    def bisect_(ds0):
        try:
            ds_, rr = bisect(func, a=0.0, b=ds0, full_output=True)
        # Will be raised when f(a) and f(b) have the same sign.
        except ValueError:
            ds_ = ds0
        return ds_

    plus_bound = prepare_bisect(0.1)
    ds_plus = bisect_(plus_bound)

    minus_bound = prepare_bisect(-0.1)
    ds_minus = bisect_(minus_bound)

    # import matplotlib.pyplot as plt
    # dss = np.linspace(-1, 1, num=51)
    # # dss = np.linspace(-6, 6, num=200)
    # E0 = E_taylor(0.0)
    # Es = E_taylor(dss) - E0
    # Es *= AU2KJPERMOL
    # Emp = (np.array((E_taylor(ds_minus), E_taylor(ds_plus))) - E0) * AU2KJPERMOL
    # _, ax = plt.subplots()
    # ax.plot(dss, Es, "o-")
    # ax.axvline(0.0, c="k", ls="--")
    # ax.axhline(dE*AU2KJPERMOL, c="k", ls=":")
    # ax.scatter((ds_minus, ds_plus), Emp, s=75, marker="x", c="r", zorder=3)
    # ax.set_xlabel("ds")
    # ax.set_ylabel("dE / kJ mol⁻¹")
    # plt.show()

    def step(ds):
        return ds * v0 + ds ** 2 * v1 / 2

    step_plus = step(ds_plus)
    step_minus = step(ds_minus)
    return step_plus, step_minus


def cubic_displ_for_h5(h5_fn="third_deriv.h5", dE=-5e-4):
    with h5py.File(h5_fn, "r") as handle:
        coords3d = handle["coords3d"][:]
        if coords3d.size > 3:
            H = handle["H_proj"][:]
        else:
            H = handle["H_mw"][:]
        Gv = handle["G_vec"][:]

    w, v = np.linalg.eigh(H)
    w0 = w[0]
    v0 = v[:, 0]
    return cubic_displ(H, v0, w0, Gv, dE)


def cubic_displ_for_geom(geom, dE=-5e-4):
    H = geom.mw_hessian
    # Only project for multi-atomic geometries.
    if geom.coords.size > 3:
        H = geom.eckart_projection(H)
    w, v = np.linalg.eigh(H)
    # Transition vector (imaginary mode) and corresponding eigenvalue
    v0 = v[:, 0]
    w0 = w[0]
    Gv, third_deriv_res = third_deriv_fd(geom, v0)
    step_plus, step_minus = cubic_displ(H, v0, w0, Gv, dE=dE)
    return step_plus, step_minus, third_deriv_res



import logging


logger = logging.getLogger("mwfn")
logger.setLevel(logging.DEBUG)
handler = logging.FileHandler("mwfn.log", mode="w", delay=True)
# fmt_str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
fmt_str = "%(asctime)s %(message)s"
formatter = logging.Formatter(fmt_str, datefmt='%Y-%m-%d %H:%M:%S')
handler.setFormatter(formatter)
logger.addHandler(handler)



import re

import numpy as np
import pyparsing as pp

from pysisyphus.elem_data import ATOMIC_NUMBERS
from pysisyphus.Geometry import Geometry
from pysisyphus.io.xyz import parse_xyz
from pysisyphus.helpers_pure import file_or_str
from pysisyphus.wavefunction import MoldenShells, Shell, Wavefunction
from pysisyphus.wavefunction.helpers import BFType


@file_or_str(".molden", ".input")
def parse_molden_atoms_coords(molden_str):
    _, geometries = re.split(r"\[GEOMETRIES\] \(XYZ\)", molden_str)
    atoms_coords, comments = parse_xyz(geometries, with_comment=True)

    return atoms_coords, comments


def geoms_from_molden(fn, **kwargs):
    atoms_coords, comments = parse_molden_atoms_coords(fn)
    geoms = [
        Geometry(atoms, coords.flatten(), comment=comment, **kwargs)
        for (atoms, coords), comment in zip(atoms_coords, comments)
    ]
    return geoms


def parse_mo(mo_lines):
    # sym, ene_key, energy, spin_key, spin, occup_key, occup, *rest = mo_lines
    sym, ene_line, spin_line, occup_line, *rest = mo_lines
    ene_key, energy = ene_line.split("=")
    spin_key, spin = spin_line.split("=")
    occup_key, occup = occup_line.split("=")
    assert (
        (ene_key.strip() == "ene")
        and (spin_key.strip() == "spin")
        and (occup_key.strip() == "occup")
    )
    prev_ind = 0
    coeffs = list()
    for line in rest:
        ind, coeff = line.split()
        ind = int(ind)
        assert ind == prev_ind + 1
        prev_ind = ind
        coeffs.append(coeff)
    return {
        "sym": sym.strip(),
        "ene": float(energy),
        "spin": spin.strip(),
        "occup": float(occup),
        "coeffs": np.array(coeffs),
    }


@file_or_str(".molden", ".input")
def parse_molden(text, with_mos=True):
    int_ = pp.common.integer
    real = pp.common.real
    sci_real = pp.common.sci_real

    def get_section(name):
        return pp.CaselessLiteral(f"[{name}]")

    molden_format = get_section("Molden Format")
    n_atoms = get_section("N_Atoms") + int_
    atoms_header = get_section("Atoms")
    title = get_section("Title") + pp.Optional(
        pp.ZeroOrMore(~atoms_header + pp.Word(pp.printables))
    )

    alpha_re = re.compile("([a-zA-Z]+)")

    def drop_number(toks):
        """OpenMolcas writes atoms like H1, C2, ...
        We don't want the number."""
        return alpha_re.match(toks[0]).group(1)

    # [Atoms]
    # element_name number atomic_number x y z
    unit = pp.one_of(("AU Angs (AU) (Angs)"), caseless=True)
    atom_symbol = pp.Word(pp.alphanums).set_parse_action(drop_number)
    xyz = pp.Group(sci_real + sci_real + sci_real)
    atom_line = pp.Group(
        atom_symbol.set_results_name("symbol")
        + int_.set_results_name("number")
        + int_.set_results_name("atomic_number")
        + xyz.set_results_name("xyz")
    )
    atoms = (
        atoms_header
        + pp.Optional(unit).set_results_name("unit")
        + pp.OneOrMore(atom_line).set_results_name("atoms")
    )

    # [Charge], written by OpenMOLCAS
    charge = get_section("Charge")
    charge_kind = pp.one_of(("(Mulliken)",), caseless=True)
    charges = charge + charge_kind + pp.OneOrMore(sci_real)

    def pgto_exps_coeffs(s, loc, toks):
        data = toks.asList()
        exps, coeffs = np.array(data, dtype=float).reshape(-1, 2).T
        pr = pp.ParseResults.from_dict(
            {
                "exponents": exps,
                "coeffs": coeffs,
            }
        )
        # Somehow this function will result in exponents and coeffs being both
        # stored in lists of length 1.
        return pr

    # [GTO]
    ang_mom = pp.one_of("s p sp d f g h i", caseless=True)
    pgto = sci_real + sci_real
    shell = pp.Group(
        ang_mom.set_results_name("ang_mom")
        + int_.set_results_name("contr_depth")
        # Usually 1.0, can be left out
        + pp.Optional(real.set_whitespace_chars(" ").suppress() + pp.LineEnd())
        + pp.OneOrMore(pgto).set_parse_action(pgto_exps_coeffs)
    )
    atom_gtos = pp.Group(
        int_.set_results_name("number")  # Center
        + pp.Optional(pp.Literal("0"))
        + pp.OneOrMore(shell).set_results_name("shells")
    )
    gto = (
        pp.CaselessLiteral("[GTO]")
        + pp.Optional(unit)
        + pp.Group(pp.OneOrMore(atom_gtos)).set_results_name("gto")
    )

    # [5D] [7F] [9G] etc.
    ang_mom_flags = pp.ZeroOrMore(pp.one_of("[5D] [7F] [9G]", caseless=True))

    # Pyparsing code to parse MOs ... slow. Regex-based code is much faster.
    # [MO]
    # sym_label = pp.Word(pp.printables)
    # spin = pp.one_of("Alpha Beta", caseless=True)
    # mo_coeff = pp.Suppress(int_) + real
    # mo = pp.Group(
    # pp.CaselessLiteral("Sym=").suppress()
    # + sym_label.set_results_name("sym")
    # + pp.CaselessLiteral("Ene=").suppress()
    # + sci_real.set_results_name("ene")
    # + pp.CaselessLiteral("Spin=").suppress()
    # + spin.set_results_name("spin")
    # + pp.CaselessLiteral("Occup=").suppress()
    # + real.set_results_name("occup")
    # + pp.Group(pp.OneOrMore(mo_coeff)).set_results_name("coeffs")
    # )
    # mos = pp.CaselessLiteral("[MO]") + pp.OneOrMore(mo).set_results_name("mos")
    mos = pp.CaselessLiteral("[MO]") + pp.OneOrMore(
        pp.Regex(r"[^\[]+")  # Match everything until the next [ is encountered
    ).set_results_name("mos")

    # Actual parser
    parser = molden_format + pp.Each(
        [
            pp.Optional(expr)
            for expr in (n_atoms, title, atoms, charges, gto, ang_mom_flags)
        ]
    )
    if with_mos:
        parser += mos

    res = parser.parse_string(text)
    as_dict = res.asDict()
    if with_mos:
        mos_raw = as_dict["mos"][0]

        by_mo = [bm.strip().split("\n") for bm in mos_raw.lower().strip().split("sym=")]
        assert by_mo[0] == [""], by_mo[0]
        by_mo = by_mo[1:]

        as_dict["mos"] = [parse_mo(mo_lines) for mo_lines in by_mo]
    return as_dict


def parse_molden_atoms(data):
    atoms = list()
    coords = list()
    nuc_charges = list()
    for atom in data["atoms"]:
        atoms.append(atom["symbol"])
        coords.extend(atom["xyz"])
        Z = atom["atomic_number"]
        nuc_charges.append(Z)
    atoms = tuple(atoms)
    coords = np.array(coords, dtype=float).flatten()
    nuc_charges = np.array(nuc_charges, dtype=int)
    return atoms, coords, nuc_charges


@file_or_str(".molden", ".input")
def shells_from_molden(text):
    data = parse_molden(text, with_mos=False)

    atoms, coords, _ = parse_molden_atoms(data)
    atomic_numbers = [ATOMIC_NUMBERS[atom.lower()] for atom in atoms]
    coords3d = coords.reshape(-1, 3)

    _shells = list()
    for atom_gtos, center, atomic_num in zip(data["gto"], coords3d, atomic_numbers):
        center_ind = atom_gtos["number"] - 1
        for shell in atom_gtos["shells"]:
            L = shell["ang_mom"]
            exps = shell["exponents"][0]
            coeffs = shell["coeffs"][0]
            shell = Shell(
                L=L,
                center=center,
                coeffs=coeffs,
                exps=exps,
                center_ind=center_ind,
                atomic_num=atomic_num,
            )
            _shells.append(shell)

    shells = MoldenShells(_shells)
    return shells


@file_or_str(".molden", ".input")
def wavefunction_from_molden(text, charge=None, shells_func=None, **wf_kwargs):
    """Construct Wavefunction object from .molden file.

    shells_func is used for ORCA, to modify the contraction coefficients.
    """
    data = parse_molden(text)
    atoms, coords, nuc_charges = parse_molden_atoms(data)
    nuc_charge = sum(nuc_charges)

    spins = list()
    occ_a = 0.0
    occ_b = 0.0
    Ca = list()
    Cb = list()
    for mo in data["mos"]:
        spin = mo["spin"].lower()
        occ = mo["occup"]
        spins.append(spin)
        coeffs = mo["coeffs"]
        if spin == "alpha":
            occ_a += occ
            Ca.append(coeffs)
        elif spin == "beta":
            occ_b += occ
            Cb.append(coeffs)
        else:
            raise Exception(f"Spin can only be 'alpha' or 'beta', but got '{spin}'!")
    assert occ_a.is_integer
    assert occ_b.is_integer
    occ_a = int(occ_a)
    occ_b = int(occ_b)

    # MOs must be in columns
    Ca = np.array(Ca, dtype=float).T
    Cb = np.array(Cb, dtype=float).T
    # Restricted calculation
    if Cb.size == 0:
        Cb = Ca.copy()
        occ_a = occ_b = occ_a // 2
    C = np.stack((Ca, Cb))

    molden_charge = nuc_charge - (occ_a + occ_b)
    if charge is None:
        charge = molden_charge
    # Multiplicity = (2S + 1), S = number of unpaired elecs. * 0.5
    mult = (occ_a - occ_b) + 1
    unrestricted = set(spins) == {"Alpha", "Beta"}

    if shells_func is None:
        shells_func = shells_from_molden
    shells = shells_func(text)

    return Wavefunction(
        atoms=atoms,
        coords=coords,
        charge=charge,
        mult=mult,
        unrestricted=unrestricted,
        occ=(occ_a, occ_b),
        C=C,
        bf_type=BFType.PURE_SPHERICAL,
        shells=shells,
        **wf_kwargs,
    )



# [1] https://link.springer.com/article/10.1007/s00214-016-1847-3
#     Birkholz, 2016
# [2] Geometry optimization in Cartesian coordinates: Constrained optimization
#     Baker, 1992
# [3] https://epubs.siam.org/doi/pdf/10.1137/S1052623496306450
#     BFGS WITH UPDATE SKIPPING AND VARYING MEMORY
#     Kolda, 1998
# [4] https://link.springer.com/article/10.1186/1029-242X-2012-241
#     New cautious BFGS algorithm based on modified Armijo-type line search
#     Wan, 2012
# [5] Numerical optimization, 2nd ed.
#     Nocedal, Wright
# [6] https://arxiv.org/abs/2006.08877
#     Goldfarb, 2020
# [7] https://pubs.acs.org/doi/10.1021/acs.jctc.9b00869
#     Hermes, Zádor, 2019
# [8] https://doi.org/10.1002/(SICI)1096-987X(199802)19:3<349::AID-JCC8>3.0.CO;2-T
#     Bofill, 1998
# [9] http://dx.doi.org/10.1016/S0166-1280(02)00209-9
#     Bungay, Poirier


import numpy as np

from pysisyphus.optimizers.closures import bfgs_multiply

import torch


def _outer(a, b):
    """Torch based outer product preserving dtype and device."""
    if isinstance(a, torch.Tensor):
        b = torch.as_tensor(b, dtype=a.dtype, device=a.device)
        return torch.outer(a, b)
    return np.outer(a, b)

def _dot(a, b):
    """Torch based dot product preserving dtype and device."""
    if isinstance(a, torch.Tensor):
        b = torch.as_tensor(b, dtype=a.dtype, device=a.device)
        return torch.dot(a, b)
    return np.dot(a, b)


def bfgs_update(H, dx, dg):
    """Standard BFGS Hessian update using torch."""
    if isinstance(H, torch.Tensor):
        dx = torch.as_tensor(dx, dtype=H.dtype, device=H.device)
        dg = torch.as_tensor(dg, dtype=H.dtype, device=H.device)

    Hdx = H @ dx
    first_term = _outer(dg, dg) / _dot(dg, dx)
    second_term = _outer(Hdx, Hdx) / _dot(dx, Hdx)
    update = first_term - second_term
    if isinstance(H, torch.Tensor):
        return update.to(dtype=H.dtype, device=H.device), "BFGS"
    return update, "BFGS"


def damped_bfgs_update(H, dx, dg):
    # if isinstance(H, torch.Tensor):
    #     dx = torch.as_tensor(dx, dtype=H.dtype, device=H.device)
    #     dg = torch.as_tensor(dg, dtype=H.dtype, device=H.device)

    #     dxdg  = _dot(dx, dg)
    #     dxHdx = _dot(dx, H @ dx)
    #     theta = 1.0
    #     if dxdg < 0.2 * dxHdx:
    #         theta = 0.8 * dxHdx / (dxHdx - dxdg)
    #     r = theta * dg + (1 - theta) * (H @ dx)

    #     first_term  = _outer(r, r) / _dot(r, dx)
    #     second_term = (H @ _outer(dx, dx) @ H) / dxHdx
    #     return first_term - second_term, "damped BFGS"

    """See [5]"""
    dxdg = dx.dot(dg)
    dxHdx = dx.dot(H).dot(dx)
    theta = 1
    if dxdg < 0.2 * dxHdx:
        theta = 0.8 * dxHdx / (dxHdx - dxdg)
    r = theta * dg + (1 - theta) * H.dot(dx)

    first_term = np.outer(r, r) / r.dot(dx)
    second_term = H.dot(np.outer(dx, dx)).dot(H) / dxHdx
    return first_term - second_term, "damped BFGS"


def double_damp(
    s, y, H=None, s_list=None, y_list=None, mu_1=0.2, mu_2=0.2, logger=None
):
    """Double damped step 's' and gradient differences 'y'.

    H is the inverse Hessian!
    See [6]. Potentially updates s and y. y is only
    updated if mu_2 is not None.

    Parameters
    ----------
    s : np.array, shape (N, ), floats
        Coordiante differences/step.
    y : np.array, shape (N, ), floats
        Gradient differences
    H : np.array, shape (N, N), floats, optional
        Inverse Hessian.
    s_list : list of nd.array, shape (K, N), optional
        List of K previous steps. If no H is supplied and prev_ys is given
        the matrix-vector product Hy will be calculated through the
        two-loop LBFGS-recursion.
    y_list : list of nd.array, shape (K, N), optional
        List of K previous gradient differences. See s_list.
    mu_1 : float, optional
        Parameter for 's' damping.
    mu_2 : float, optional
        Parameter for 'y' damping.
    logger : logging.Logger, optional
        Logger to be used.

    Returns
    -------
    s : np.array, shape (N, ), floats
        Damped coordiante differences/step.
    y : np.array, shape (N, ), floats
        Damped gradient differences
    """
    # if isinstance(H, torch.Tensor):
    #     s = torch.as_tensor(s, dtype=H.dtype, device=H.device)
    #     y = torch.as_tensor(y, dtype=H.dtype, device=H.device)

    #     sy  = _dot(s, y)
    #     Hy  = H @ y
    #     yHy = _dot(y, Hy)

    #     theta_1 = 1.0
    #     if sy < mu_1 * yHy:
    #         theta_1 = (1 - mu_1) * yHy / (yHy - sy)
    #         s = theta_1 * s + (1 - theta_1) * Hy

    #     # --- 二重 damping ---
    #     sy   = _dot(s, y)
    #     ss   = _dot(s, s)
    #     theta_2 = 1.0
    #     if mu_2 is not None and sy < mu_2 * ss:
    #         theta_2 = (1 - mu_2) * ss / (ss - sy)
    #         y = theta_2 * y + (1 - theta_2) * s

    #     return s, y

    sy = s.dot(y)
    # Calculate Hy directly
    if H is not None:
        Hy = H.dot(y)
    # Calculate Hy via BFGS_multiply as in LBFGS
    else:
        Hy = bfgs_multiply(s_list, y_list, y, logger=logger)
    yHy = y.dot(Hy)

    theta_1 = 1
    damped_s = ""
    if sy < mu_1 * yHy:
        theta_1 = (1 - mu_1) * yHy / (yHy - sy)
        s = theta_1 * s + (1 - theta_1) * Hy
        if theta_1 < 1.0:
            damped_s = ", damped s"
    msg = f"damped BFGS\n\ttheta_1={theta_1:.4f} {damped_s}"

    # Double damping
    damped_y = ""
    if mu_2 is not None:
        sy = s.dot(y)
        ss = s.dot(s)
        theta_2 = 1
        if sy < mu_2 * ss:
            theta_2 = (1 - mu_2) * ss / (ss - sy)
        y = theta_2 * y + (1 - theta_2) * s
        if theta_2 < 1.0:
            damped_y = ", damped y"
        msg = "double " + msg + f"\n\ttheta_2={theta_2:.4f} {damped_y}"

    if logger is not None:
        logger.debug(msg.capitalize())

    return s, y


def sr1_update(z, dx):
    """Symmetric rank-one update."""
    if isinstance(z, torch.Tensor):
        dx = torch.as_tensor(dx, dtype=z.dtype, device=z.device)
    update = _outer(z, z) / _dot(z, dx)
    if isinstance(z, torch.Tensor):
        return update.to(dtype=z.dtype, device=z.device), "SR1"
    return update, "SR1"


def psb_update(z, dx):
    """Powell Symmetric Broyden update."""
    if isinstance(z, torch.Tensor):
        dx = torch.as_tensor(dx, dtype=z.dtype, device=z.device)
    first = (_outer(dx, z) + _outer(z, dx)) / _dot(dx, dx)
    second = _dot(dx, z) * _outer(dx, dx) / (_dot(dx, dx) ** 2)
    update = first - second
    if isinstance(z, torch.Tensor):
        return update.to(dtype=z.dtype, device=z.device), "PSB"
    return update, "PSB"


def flowchart_update(H, dx, dg):
    # if isinstance(H, torch.Tensor):
    #     dx = torch.as_tensor(dx, dtype=H.dtype, device=H.device)
    #     dg = torch.as_tensor(dg, dtype=H.dtype, device=H.device)

    #     z          = dg - H @ dx
    #     sr1_quot   = _dot(z, dx) / (z.norm() * dx.norm())
    #     bfgs_quot  = _dot(dg, dx) / (dg.norm() * dx.norm())

    #     if sr1_quot < -0.1:
    #         update, key = sr1_update(z, dx)
    #     elif bfgs_quot > 0.1:
    #         update, key = bfgs_update(H, dx, dg)
    #     else:
    #         update, key = psb_update(z, dx)
    #     return update, key

    # See [1], Sec. 2, equations 1 to 3
    z = dg - H.dot(dx)
    sr1_quot = z.dot(dx) / (np.linalg.norm(z) * np.linalg.norm(dx))
    bfgs_quot = dg.dot(dx) / (np.linalg.norm(dg) * np.linalg.norm(dx))
    if sr1_quot < -0.1:
        update, key = sr1_update(z, dx)
    elif bfgs_quot > 0.1:
        update, key = bfgs_update(H, dx, dg)
    else:
        update, key = psb_update(z, dx)
    return update, key


def mod_flowchart_update(H, dx, dg):
    # This version seems to work too ... at least for minimizations
    # starting from a geometry near a transition state. Interesing.
    z = dg - H.dot(dx)
    quot = z.dot(dx) / (np.linalg.norm(z) * np.linalg.norm(dx))
    if quot < -0.1:
        update, key = bfgs_update(H, dx, dg)
    elif quot > 0.1:
        update, key = sr1_update(z, dx)
    else:
        update, key = psb_update(z, dx)
    return update, key


def bofill_update(H, dx, dg):
    """Bofill's combination of SR1 and PSB updates."""
    if isinstance(H, torch.Tensor):
        dx = torch.as_tensor(dx, dtype=H.dtype, device=H.device)
        dg = torch.as_tensor(dg, dtype=H.dtype, device=H.device)

    z = dg - H @ dx
    sr1 = sr1_update(z, dx)[0]
    psb = psb_update(z, dx)[0]
    mix = (_dot(z, dx) ** 2) / (_dot(z, z) * _dot(dx, dx))
    update = mix * sr1 + (1 - mix) * psb
    if isinstance(H, torch.Tensor):
        return update.to(dtype=H.dtype, device=H.device), "Bofill"
    return update, "Bofill"


def ts_bfgs_update(H, dx, dg):
    # if isinstance(H, torch.Tensor):
    #     dx = torch.as_tensor(dx, dtype=H.dtype, device=H.device)[:, None]
    #     dg = torch.as_tensor(dg, dtype=H.dtype, device=H.device)[:, None]
    #     j   = dg - H @ dx
    #     jdx = (j.T @ dx)

    #     w, v = torch.linalg.eigh(H)
    #     Hdx  = torch.abs(w) * (v @ (v.T @ dx))
    #     M     = dg @ dg.T + Hdx @ Hdx.T
    #     dxTM  = dx.T @ M
    #     u     = torch.linalg.solve(dxTM @ dx, dxTM).T
    #     juT   = j @ u.T
    #     update = juT + juT.T - jdx * (u @ u.T)
    #     return update, "TS-BFGS"

    """As described in [7]"""
    dx = dx[:, None]
    dg = dg[:, None]
    j = dg - H @ dx
    jdx = j.T @ dx
    # Diagonalize Hessian, to construct positive definite version of it
    w, v = np.linalg.eigh(H)
    Hdx = np.abs(w) * v @ (v.T @ dx)
    M = dg @ dg.T + Hdx @ Hdx.T
    dxTM = dx.T @ M
    u = np.linalg.solve(dxTM @ dx, dxTM).T
    juT = j @ u.T
    ts_bfgs_update = juT + juT.T - jdx * u @ u.T
    return ts_bfgs_update, "TS-BFGS"


def ts_bfgs_update_org(H, dx, dg):
    """Do not use! Implemented as described in the 1998 bofill paper [8].

    This does not seem to work too well."""
    dx = dx[:, None]
    dg = dg[:, None]
    u = dg
    j = H @ dx
    j = u - j
    # jTdx = float(j.T @ dx)
    jTdx = j.T @ dx
    # dxTdx = float(dx.T @ dx)
    dxTdx = dx.T @ dx
    # jTj = float(j.T @ j)
    jTj = j.T @ j
    phi = jTdx ** 2 / dxTdx / jTj
    u = phi * dg * dg.T @ dx
    w, v = np.linalg.eigh(H)
    u = u + (1 - phi) * np.abs(w) * v @ (v.T @ dx)
    u = u / (u.T @ dx)
    juT = j @ u.T
    ts_bfgs_update = juT + juT.T - jTdx * u @ u.T
    return ts_bfgs_update, "TS-BFGS"


def ts_bfgs_update_revised(H, dx, dg):
    """TS-BFGS update as described in [9].

    Better than the original formula of Bofill, worse than the implementation
    in [7]. a is caluclated as described in the footnote 1 on page 38. Eq. (8)
    looks suspicious as it contains the inverse of a vector?! As also outlined
    in the paper abs(a) is used (|a| in the paper)."""

    dx = dx[:, None]
    dg = dg[:, None]
    dgTdg = dg.T @ dg
    dgTdx = dg.T @ dx
    a = (dgTdg - dg.T @ H @ dx) / (dgTdg * dgTdx)
    a = abs(a)

    # Diagonalize Hessian, to construct positive definite version of it
    w, v = np.linalg.eigh(H)
    H_pos_dx = np.abs(w) * v @ (v.T @ dx)
    # Mixing factor
    j = dg - H @ dx
    jTdx = j.T @ dx
    dxTdx = dx.T @ dx
    jTj = j.T @ j
    phi = jTdx ** 2 / dxTdx / jTj
    u = ((1 - phi) * H_pos_dx + a * phi * dgTdx * dg) / (
        (1 - phi) * dx.T @ H_pos_dx + phi * a * dgTdx ** 2
    )

    juT = j @ u.T
    ts_bfgs_update = juT + juT.T - jTdx * u @ u.T
    return ts_bfgs_update, "TS-BFGS"



import numpy as np

from pysisyphus.intcoords.Primitive import Primitive
from pysisyphus.intcoords.derivatives import dq_ld, d2q_ld
from pysisyphus.linalg import cross3, norm3


class LinearDisplacement(Primitive):
    def __init__(self, *args, complement=False, **kwargs):
        kwargs["calc_kwargs"] = ("complement", "cross_vec")
        super().__init__(*args, **kwargs)

        self.complement = complement
        self.cross_vec = None

    @staticmethod
    def _weight(atoms, coords3d, indices, f_damping):
        raise Exception("Not yet implemented!")

    def calculate(self, coords3d, indices=None, gradient=False):
        if self.cross_vec is None:
            self.set_cross_vec(coords3d, indices)

        return super().calculate(coords3d, indices, gradient)

    @staticmethod
    def _calculate(coords3d, indices, gradient=False, complement=False, cross_vec=None):
        m, o, n = indices
        w_dash = coords3d[n] - coords3d[m]
        w = w_dash / norm3(w_dash)

        u_dash = coords3d[m] - coords3d[o]
        v_dash = coords3d[n] - coords3d[o]
        u = u_dash / norm3(u_dash)
        v = v_dash / norm3(v_dash)

        # Vector for cross product to determine first orthogonal direction
        if cross_vec is None:
            cross_vec = LinearDisplacement._get_cross_vec(coords3d, indices)

        if complement:
            cross_vec = cross3(w, cross_vec)
        cross_vec /= norm3(cross_vec)

        # Orthogonal direction
        y = cross3(w, cross_vec)
        y /= norm3(y)

        lin_disp = y.dot(u) + y.dot(v)

        if gradient:
            row = np.zeros_like(coords3d)
            row[indices] = dq_ld(*coords3d[indices].flatten(), *cross_vec).reshape(-1, 3)
            return lin_disp, row.flatten()

        return lin_disp

    def jacobian(self, coords3d, indices=None):
        if self.cross_vec is None:
            self.set_cross_vec(coords3d, indices)

        return super().jacobian(coords3d, indices)

    @staticmethod
    def _jacobian(coords3d, indices, complement=False, cross_vec=None):
        if cross_vec is None:
            cross_vec = LinearDisplacement._get_cross_vec(coords3d, indices)

        if complement:
            m, _, n = indices
            w_dash = coords3d[n] - coords3d[m]
            w = w_dash / norm3(w_dash)
            cross_vec = cross3(w, cross_vec)

        return d2q_ld(*coords3d[indices].flatten(), *cross_vec)

    def __str__(self):
        return (
            f"LinearDisplacement({tuple(self.indices)}, complement={self.complement})"
        )



import itertools as it
from typing import Union, Sequence

import numpy as np

from pysisyphus.constants import BOHR2ANG
from pysisyphus.helpers_pure import OrderedEnum
from pysisyphus.intcoords import (
    Bend,
    Bend2,
    BondedFragment,
    DummyImproper,
    DummyTorsion,
    DistanceFunction,
    CartesianX,
    CartesianY,
    CartesianZ,
    LinearBend,
    LinearDisplacement,
    OutOfPlane,
    RobustTorsion1,
    RobustTorsion2,
    RotationA,
    RotationB,
    RotationC,
    Stretch,
    TranslationX,
    TranslationY,
    TranslationZ,
    Torsion,
    Torsion2,
)


class PrimTypes(OrderedEnum):
    BOND = 0
    AUX_BOND = 1
    HYDROGEN_BOND = 2
    INTERFRAG_BOND = 3
    AUX_INTERFRAG_BOND = 4
    BEND = 5
    LINEAR_BEND = 6
    LINEAR_BEND_COMPLEMENT = 7
    PROPER_DIHEDRAL = 8
    IMPROPER_DIHEDRAL = 9
    OUT_OF_PLANE = 10
    LINEAR_DISPLACEMENT = 11
    LINEAR_DISPLACEMENT_COMPLEMENT = 12
    TRANSLATION = 13
    TRANSLATION_X = 14
    TRANSLATION_Y = 15
    TRANSLATION_Z = 16
    ROTATION = 17
    ROTATION_A = 18
    ROTATION_B = 19
    ROTATION_C = 20
    CARTESIAN = 21
    CARTESIAN_X = 22
    CARTESIAN_Y = 23
    CARTESIAN_Z = 24
    BONDED_FRAGMENT = 25
    DUMMY_TORSION = 26
    DISTANCE_FUNCTION = 27
    # atan2 based bend and torsion
    BEND2 = 28
    PROPER_DIHEDRAL2 = 29
    DUMMY_IMPROPER = 30
    ROBUST_TORSION1 = 31
    ROBUST_TORSION2 = 32


PrimTypeLike = Union[PrimTypes, str]


# Alias for easier access
PT = PrimTypes


PrimTypeShortcuts = {
    "X": [PT.CARTESIAN_X],
    "Y": [PT.CARTESIAN_Y],
    "Z": [PT.CARTESIAN_Z],
    "XY": [PT.CARTESIAN_X, PT.CARTESIAN_Y],
    "XZ": [PT.CARTESIAN_X, PT.CARTESIAN_Z],
    "YZ": [PT.CARTESIAN_Y, PT.CARTESIAN_Z],
    "XYZ": [PT.CARTESIAN_X, PT.CARTESIAN_Y, PT.CARTESIAN_Z],
    "ATOM": [PT.CARTESIAN_X, PT.CARTESIAN_Y, PT.CARTESIAN_Z],
    # Primitive aliases
    "B": [PT.BOND],
    "A": [PT.BEND],
    "A2": [PT.BEND2],
    "D": [PT.PROPER_DIHEDRAL],
    "D2": [PT.PROPER_DIHEDRAL2],
    "DIHEDRAL": [PT.PROPER_DIHEDRAL],
    "DIHEDRAL2": [PT.PROPER_DIHEDRAL2],
    "TORSION": [PT.PROPER_DIHEDRAL],
    "TORSION2": [PT.PROPER_DIHEDRAL2],
    "RTORSION1": [PT.ROBUST_TORSION1],
    "RTORSION2": [PT.ROBUST_TORSION2],
    # Translation & Rotation coordinates
    "TRANSLATION": [PT.TRANSLATION_X, PT.TRANSLATION_Y, PT.TRANSLATION_Z],
    "ROTATION": [PT.ROTATION_A, PT.ROTATION_B, PT.ROTATION_C],
    "DIST_FUNC": [PT.DISTANCE_FUNCTION],
}

# The tuples below can be used to decide whether a given type belongs
# to a certain class of primitive.
Bonds = (
    PT.BOND,
    PT.AUX_BOND,
    PT.HYDROGEN_BOND,
    PT.INTERFRAG_BOND,
    PT.AUX_INTERFRAG_BOND,
    PT.DISTANCE_FUNCTION,
)
Bends = (PT.BEND, PT.BEND2)
LinearBends = (
    PT.LINEAR_BEND,
    PT.LINEAR_BEND_COMPLEMENT,
    PT.LINEAR_DISPLACEMENT,
    PT.LINEAR_DISPLACEMENT_COMPLEMENT,
)
Dihedrals = (
    PT.PROPER_DIHEDRAL,
    PT.IMPROPER_DIHEDRAL,
    PT.PROPER_DIHEDRAL2,
    PT.DUMMY_IMPROPER,
    PT.ROBUST_TORSION1,
    PT.ROBUST_TORSION2,
)
OutOfPlanes = (PT.OUT_OF_PLANE,)
Cartesians = (PT.CARTESIAN_X, PT.CARTESIAN_Y, PT.CARTESIAN_Z)
Rotations = (PT.ROTATION_A, PT.ROTATION_B, PT.ROTATION_C)
Translations = (
    PT.TRANSLATION_X,
    PT.TRANSLATION_Y,
    PT.TRANSLATION_Z,
    PT.BONDED_FRAGMENT,
)
DummyCoords = (PT.DUMMY_TORSION,)


def get_rot_coord(cls):
    def func(indices, ref_coords3d):
        return cls(indices, ref_coords3d=ref_coords3d)

    return func


def get_bonded_frag_coord():
    def func(indices):
        indices_ = indices[:-2]
        bond_indices = indices[-2:]
        return BondedFragment(indices_, bond_indices=bond_indices)

    return func


def get_dist_func():
    def func(indices):
        indices_ = indices[:4]
        coeff = indices[4]
        return DistanceFunction(indices_, coeff=coeff)

    return func


# Maps primitive types to their classes
PrimMap = {
    PT.BOND: Stretch,
    PT.AUX_BOND: Stretch,
    PT.HYDROGEN_BOND: Stretch,
    PT.INTERFRAG_BOND: Stretch,
    PT.AUX_INTERFRAG_BOND: Stretch,
    PT.BEND: Bend,
    PT.BEND2: Bend2,
    PT.LINEAR_BEND: LinearBend,
    PT.LINEAR_BEND_COMPLEMENT: lambda indices: LinearBend(indices, complement=True),
    PT.PROPER_DIHEDRAL: lambda indices: Torsion(indices, periodic=True),
    PT.PROPER_DIHEDRAL2: lambda indices: Torsion2(indices, periodic=True),
    PT.IMPROPER_DIHEDRAL: lambda indices: Torsion(indices, periodic=True),
    PT.ROBUST_TORSION1: lambda indices: RobustTorsion1(indices),
    PT.ROBUST_TORSION2: lambda indices: RobustTorsion2(indices),
    PT.OUT_OF_PLANE: OutOfPlane,
    PT.LINEAR_DISPLACEMENT: LinearDisplacement,
    PT.LINEAR_DISPLACEMENT_COMPLEMENT: lambda indices: LinearDisplacement(
        indices, complement=True
    ),
    PT.TRANSLATION_X: TranslationX,
    PT.TRANSLATION_Y: TranslationY,
    PT.TRANSLATION_Z: TranslationZ,
    PT.ROTATION_A: get_rot_coord(RotationA),
    PT.ROTATION_B: get_rot_coord(RotationB),
    PT.ROTATION_C: get_rot_coord(RotationC),
    PT.CARTESIAN_X: CartesianX,
    PT.CARTESIAN_Y: CartesianY,
    PT.CARTESIAN_Z: CartesianZ,
    PT.BONDED_FRAGMENT: get_bonded_frag_coord(),
    PT.DUMMY_TORSION: lambda indices: DummyTorsion(
        indices,
        periodic=True,
    ),
    PT.DISTANCE_FUNCTION: get_dist_func(),
    PT.DUMMY_IMPROPER: lambda indices: DummyImproper(
        indices,
        periodic=True,
    ),
}


def normalize_prim_input(prim_inp):
    """Normalize input for define_prims and constrain_prims

    The intcoords.RedundantCoords constructor expects lists of integer lists
    (tuples) for arguments like 'define_prims' and 'constrain_prims'. The first item
    of every list determines the type of primitive coordinate. Currently
    there are about 20 different types and it is hard to remember all of
    them.

    So we also allow a more human friendly input, that is normalized here.
    The most common primitives are:

    0: BOND
    5: BEND
    8: PROPER_DIHEDRAL

    This function maps inputs like ["BOND", 1, 2] to [PrimTypes.BOND, 1, 2] etc.

    Always returns a list of tuples, as some prim_inps expand to multiple
    coordinates, e.g., XYZ or ATOM.
    """
    if prim_inp is None:
        return []

    prim_type, *indices = prim_inp
    indices = list(map(int, indices))

    # Nothing to do
    if isinstance(prim_type, PrimTypes):
        return [prim_inp]

    # First check if we got something like an integer
    try:
        return [tuple([PrimTypes(int(prim_type))] + indices)]
    # Raised when prim_type is, e.g., "BOND"
    except ValueError:
        pass

    # Check if we got a PrimType name
    try:
        prim_type_ = getattr(PrimTypes, str(prim_type).upper())
        return [tuple([prim_type_] + indices)]
    except AttributeError:
        pass

    # Check if we got a shortcut, e.g, X/Y/Z/XYZ/ATOM etc.
    try:
        prim_types_ = PrimTypeShortcuts[str(prim_type).upper()]
        return [tuple([prim_type_] + indices) for prim_type_ in prim_types_]
    except KeyError as error:
        print(f"Could not normalize 'prim_inp'={prim_inp}!")
        raise error


def normalize_prim_inputs(prim_inps):
    # Flatten list of tuples
    return list(it.chain(*[normalize_prim_input(pi) for pi in prim_inps]))


def prims_from_prim_inputs(prim_inps):
    norm_prim_inps = normalize_prim_inputs(prim_inps)
    prims = [PrimMap[prim_type](indices) for prim_type, *indices in norm_prim_inps]
    return prims


def prim_for_human(prim_type: PrimTypes, val: Sequence[int]):
    if prim_type in Bonds:
        val_conv = val * BOHR2ANG
        unit = " Å"
    elif prim_type in Bends:
        unit = "°"
        val_conv = np.rad2deg(val)
    else:
        val_conv = val
        unit = ""
    return val_conv, unit



__all__ = [
    "PrimitiveNotDefinedException",
    "Bend",
    "Bend2",
    "CartesianX",
    "CartesianY",
    "CartesianZ",
    "DummyImproper",
    "DummyTorsion",
    "DistanceFunction",
    "LinearBend",
    "LinearDisplacement",
    "OutOfPlane",
    "Stretch",
    "Torsion",
    "Torsion2",
    "RobustTorsion1",
    "RobustTorsion2",
    "RotationA",
    "RotationB",
    "RotationC",
    "TranslationX",
    "TranslationY",
    "TranslationZ",
    "DLC",
    "HDLC",
    "CartesianCoords",
    "MWCartesianCoords",
    "RedundantCoords",
    "TRIC",
    "TMTRIC",
    "HybridRedundantCoords",
]

from pysisyphus.intcoords.exceptions import PrimitiveNotDefinedException
from pysisyphus.intcoords.Bend import Bend
from pysisyphus.intcoords.Bend2 import Bend2
from pysisyphus.intcoords.BondedFragment import BondedFragment
from pysisyphus.intcoords.Cartesian import CartesianX, CartesianY, CartesianZ
from pysisyphus.intcoords.DistanceFunction import DistanceFunction
from pysisyphus.intcoords.DummyImproper import DummyImproper
from pysisyphus.intcoords.DummyTorsion import DummyTorsion
from pysisyphus.intcoords.CartesianCoords import CartesianCoords, MWCartesianCoords
from pysisyphus.intcoords.LinearBend import LinearBend
from pysisyphus.intcoords.LinearDisplacement import LinearDisplacement
from pysisyphus.intcoords.OutOfPlane import OutOfPlane
from pysisyphus.intcoords.Rotation import RotationA, RotationB, RotationC
from pysisyphus.intcoords.RobustTorsion import RobustTorsion1, RobustTorsion2
from pysisyphus.intcoords.Stretch import Stretch
from pysisyphus.intcoords.Torsion import Torsion
from pysisyphus.intcoords.Torsion2 import Torsion2
from pysisyphus.intcoords.Translation import TranslationX, TranslationY, TranslationZ
from pysisyphus.intcoords.RedundantCoords import (
    RedundantCoords,
    TRIC,
    TMTRIC,
    HybridRedundantCoords,
)

# DLC inherits from RedundantCoords, so we import it after RedundantCoords
from pysisyphus.intcoords.DLC import DLC, HDLC



import sys
import time

from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.calculators import ORCA, ORCA5, Turbomole, XTB

CALC_CLASSES = {
    "orca": ORCA.ORCA,
    "orca5": ORCA5.ORCA5,
    "turbomole": Turbomole.Turbomole,
    "xtb": XTB.XTB,
}

try:
    from pysisyphus.calculators import PySCF

    CALC_CLASSES["pyscf"] = PySCF.PySCF
except ModuleNotFoundError:
    pass


def calcs_from_dict(calc_dict, base_name, calc_number, charge, mult, pal, mem):
    keys_calcs = {}
    calc_kwargs_ = {
        "calc_number": calc_number,
        "charge": charge,
        "mult": mult,
        "pal": pal,
        "mem": mem,
    }
    # Try to distinguish the calculators according to their base_name. If no
    # base_name is given we rely on the calc_number.
    base_name = base_name if base_name != "" else f"{calc_number:03d}"
    for i, (key, kwargs) in enumerate(calc_dict.items()):
        calc_kwargs = calc_kwargs_.copy()
        calc_kwargs["base_name"] = f"{base_name}_{key}"
        # Don't modify original dict
        kwargs = kwargs.copy()
        type_ = kwargs.pop("type")
        calc_kwargs.update(**kwargs)
        calc_cls = CALC_CLASSES[type_]
        keys_calcs[key] = calc_cls(**calc_kwargs)
    return keys_calcs


class MultiCalc(Calculator):
    def __init__(self, calcs, **kwargs):
        super().__init__(**kwargs)

        self.do_hess = {
            key: calc_dict.pop("do_hess", False) for key, calc_dict in calcs.items()
        }
        self.keys_calcs = calcs_from_dict(
            calcs,
            self.base_name,
            self.calc_number,
            self.charge,
            self.mult,
            self.pal,
            self.mem,
        )
        max_len = max([len(key) for key in self.keys_calcs.keys()])
        self.max_fmt = f" >{max_len+2}"

    def run_calculation(self, atoms, coords, **prepare_kwargs):
        all_results = {}
        for name, calc in self.keys_calcs.items():
            if self.do_hess[name]:
                run_func = calc.get_hessian
            else:
                run_func = calc.run_calculation
            key = f"{self.base_name}_{name}"
            print(f"Running {name:{self.max_fmt}} ... ", end="")
            try:
                start = time.time()
                results = run_func(atoms, coords, **prepare_kwargs)
                end = time.time()
                duration = end - start
                print(f"took {duration:.0f} s.")
            except Exception as err:
                print("\nCalculation failed!")
                print(err)
            sys.stdout.flush()
            all_results[key] = results
        return all_results



import logging

__all__ = [
    "BFGS",
    "ConjugateGradient",
    "CubicNewton",
    "FIRE",
    "LayerOpt",
    "LBFGS",
    "MicroOptimizer",
    "NCOptimizer",
    "PreconLBFGS",
    "PreconSteepestDescent",
    "QuickMin",
    "RFOptimizer",
    "SteepestDescent",
    "StringOptimizer",
    "StabilizedQNMethod",
]

from pysisyphus.optimizers.CubicNewton import CubicNewton
from pysisyphus.optimizers.MicroOptimizer import MicroOptimizer

logger = logging.getLogger("optimizer")
logger.setLevel(logging.DEBUG)
# delay = True prevents creation of empty logfiles
handler = logging.FileHandler("optimizer.log", mode="w", delay=True)
# fmt_str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
fmt_str = "%(message)s"
formatter = logging.Formatter(fmt_str)
handler.setFormatter(formatter)
logger.addHandler(handler)

# logger = logging.getLogger("gdiis")
# logger.setLevel(logging.DEBUG)
# # delay = True prevents creation of empty logfiles
# handler = logging.FileHandler("gdiis.log", mode="w", delay=True)
# fmt_str = "%(message)s"
# formatter = logging.Formatter(fmt_str)
# handler.setFormatter(formatter)
# logger.addHandler(handler)



# pdb2reaction/scan2d.py

"""
2D grid scan with harmonic restraints on two inter-atomic distances.

Example:
    pdb2reaction scan2d -i input.pdb -q 0 --scan-lists '[(12,45,1.30,3.10),(10,55,1.20,3.20)]'

For detailed documentation, see: docs/scan2d.md
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import logging
import math
import sys
import textwrap
import traceback
import shutil
import tempfile
import time

import click
import numpy as np
import pandas as pd
from scipy.interpolate import Rbf
import plotly.graph_objects as go

from pysisyphus.helpers import geom_loader
from pysisyphus.optimizers.exceptions import OptimizationError, ZeroStepLength
from pysisyphus.constants import ANG2BOHR, AU2KCALPERMOL

from .defaults import (
    CALC_KW_DEFAULT,
    GEOM_KW_DEFAULT,
    BIAS_KW,
    OPT_MODE_ALIASES,
    OPT_BASE_KW,
    LBFGS_KW,
    RFO_KW,
    UMA_CALC_KW,
    OUT_DIR_SCAN2D,
)
from .backends import create_calculator
from .opt import HarmonicBiasCalculator
from .utils import (
    axis_label_csv,
    axis_label_html,
    build_sopt_kwargs,
    is_scan_spec_file,
    make_sopt_optimizer,
    parse_scan_list_quads_checked,
    parse_scan_spec_quads,
    unbiased_energy_hartree,
    values_from_bounds,
    pretty_block,
    strip_inherited_keys,
    format_geom_for_echo,
    format_elapsed,
    normalize_choice,
    prepared_cli_input,
    set_convert_file_enabled,
    load_pdb_atom_metadata,
    format_pdb_atom_metadata,
    format_pdb_atom_metadata_header,
    ensure_dir,
    make_snapshot_geometry,
    convert_xyz_like_outputs,
    build_scan_configs,
    cli_param_overridden,
    resolve_freeze_atoms,
    distance_A_from_coords,
    distance_tag,
    set_freeze_atoms_or_warn,
)
from .scan_common import (
    add_scan_common_options,
    load_merged_yaml_cfg,
    resolve_yaml_sources,
)

logger = logging.getLogger(__name__)

# Defaults imported from defaults.py
DEFAULT_THRESH_2D = "baker"

_snapshot_geometry = make_snapshot_geometry(GEOM_KW_DEFAULT["coord_type"])


def _sort_values_by_reference(values: np.ndarray, ref: Optional[float]) -> np.ndarray:
    """Sort scan values so that those closest to ref come first."""
    if ref is None or not np.isfinite(ref):
        return values
    order = np.argsort(np.abs(values - ref))
    return values[order]


def _build_scan_context(
    *,
    yaml_cfg: Dict[str, Any],
    geom_kw: Dict[str, Any],
    calc_kw: Dict[str, Any],
    opt_kw: Dict[str, Any],
    lbfgs_kw: Dict[str, Any],
    rfo_kw: Dict[str, Any],
    bias_kw: Dict[str, Any],
    charge: Optional[int],
    spin: Optional[int],
    workers: int,
    workers_per_node: int,
    out_dir: str,
    thresh: Optional[str],
    bias_k: float,
    opt_mode: str,
    relax_max_cycles: int,
    relax_override_requested: bool,
    max_step_size: float,
    source_path: Optional[Path],
    freeze_links: bool,
    set_charge_spin: bool = True,
) -> Tuple[
    Dict[str, Any],
    Dict[str, Any],
    Dict[str, Any],
    Dict[str, Any],
    Dict[str, Any],
    Dict[str, Any],
    str,
    List[int],
    Path,
]:
    geom_cfg, calc_cfg, opt_cfg, lbfgs_cfg, rfo_cfg, bias_cfg = build_scan_configs(
        yaml_cfg,
        geom_kw=geom_kw,
        calc_kw=calc_kw,
        opt_kw=opt_kw,
        lbfgs_kw=lbfgs_kw,
        rfo_kw=rfo_kw,
        bias_kw=bias_kw,
        charge=charge,
        spin=spin,
        workers=workers,
        workers_per_node=workers_per_node,
        out_dir=out_dir,
        thresh=thresh,
        bias_k=bias_k,
        set_charge_spin=set_charge_spin,
    )

    kind = normalize_choice(
        opt_mode,
        param="--opt-mode",
        alias_groups=OPT_MODE_ALIASES,
        allowed_hint="grad|hess",
    )

    freeze: List[int] = []
    if source_path is not None:
        freeze = resolve_freeze_atoms(geom_cfg, source_path, freeze_links)
    calc_cfg["freeze_atoms"] = list(geom_cfg.get("freeze_atoms", []))
    calc_cfg["return_partial_hessian"] = True

    out_dir_path = Path(opt_cfg["out_dir"]).resolve()
    ensure_dir(out_dir_path)
    echo_geom = format_geom_for_echo(geom_cfg)
    echo_calc = format_geom_for_echo(calc_cfg)
    echo_opt = dict(opt_cfg)
    if relax_override_requested:
        echo_opt["max_cycles"] = int(relax_max_cycles)
    echo_opt["out_dir"] = str(out_dir_path)
    echo_bias = dict(bias_cfg)
    click.echo(pretty_block("geom", echo_geom))
    click.echo(pretty_block("calc", echo_calc))
    click.echo(pretty_block("opt", echo_opt))
    max_step_bohr_for_log = float(max_step_size) * ANG2BOHR
    echo_sopt = build_sopt_kwargs(
        kind,
        lbfgs_cfg,
        rfo_cfg,
        opt_cfg,
        max_step_bohr_for_log,
        relax_max_cycles,
        relax_override_requested,
        out_dir_path,
        str(opt_cfg.get("prefix", "")),
    )
    echo_sopt = strip_inherited_keys(echo_sopt, opt_cfg)
    click.echo(
        pretty_block("lbfgs" if kind == "lbfgs" else "rfo", echo_sopt)
    )
    click.echo(pretty_block("bias", echo_bias))

    return (
        geom_cfg,
        calc_cfg,
        opt_cfg,
        lbfgs_cfg,
        rfo_cfg,
        bias_cfg,
        kind,
        freeze,
        out_dir_path,
    )


@click.command(
    help="2D distance scan with harmonic restraints.",
    context_settings={"help_option_names": ["-h", "--help"]},
)
@click.option(
    "-i",
    "--input",
    "input_path",
    type=click.Path(path_type=Path, exists=True, dir_okay=False),
    required=True,
    help="Input structure file (.pdb, .xyz, _trj.xyz, ...).",
)
@click.option(
    "-s", "--scan-lists",
    "scan_list_raw",
    type=str,
    required=False,
    help="Scan targets: inline Python literal or a YAML/JSON spec file path.",
)
@add_scan_common_options(
    workers_default=UMA_CALC_KW["workers"],
    workers_per_node_default=UMA_CALC_KW["workers_per_node"],
    out_dir_default=OUT_DIR_SCAN2D,
    baseline_help="Reference for relative energy (kcal/mol): 'min' or 'first' (i=0,j=0).",
    dump_help="Write inner scan trajectories per d1-step as TRJ under result_scan2d/grid/.",
)
@click.option(
    "--print-parsed/--no-print-parsed",
    "print_parsed",
    default=False,
    show_default=True,
    help="Print parsed scan targets after resolving --scan-lists.",
)
@click.pass_context
def cli(
    ctx: click.Context,
    input_path: Path,
    charge: Optional[int],
    ligand_charge: Optional[str],
    workers: int,
    workers_per_node: int,
    spin: Optional[int],
    scan_list_raw: Optional[str],
    one_based: bool,
    max_step_size: float,
    bias_k: float,
    relax_max_cycles: int,
    opt_mode: str,
    freeze_links: bool,
    dump: bool,
    convert_files: bool,
    ref_pdb: Optional[Path],
    out_dir: str,
    thresh: Optional[str],
    config_yaml: Optional[Path],
    preopt: bool,
    print_parsed: bool,
    baseline: str,
    zmin: Optional[float],
    zmax: Optional[float],
    backend: str,
    solvent: str,
    solvent_model: str,
) -> None:

    set_convert_file_enabled(convert_files)
    config_yaml, override_yaml, used_legacy_yaml = resolve_yaml_sources(
        config_yaml=config_yaml,
        override_yaml=None,
        args_yaml_legacy=None,
    )

    cycles_overridden = cli_param_overridden(ctx, "relax_max_cycles")

    with prepared_cli_input(
        input_path,
        ref_pdb=ref_pdb,
        charge=charge,
        spin=spin,
        ligand_charge=ligand_charge,
        prefix="[scan2d]",
    ) as (prepared_input, resolved_charge, resolved_spin):
        geom_input_path = prepared_input.geom_path
        source_path = prepared_input.source_path

        tmp_root = None
        try:
            time_start = time.perf_counter()

            yaml_cfg = load_merged_yaml_cfg(
                config_yaml=config_yaml,
                override_yaml=None,
            )
            yaml_opt = yaml_cfg.get("opt") if isinstance(yaml_cfg, dict) else None
            relax_override_requested = cycles_overridden and not (
                isinstance(yaml_opt, dict) and "max_cycles" in yaml_opt
            )

            (
                geom_cfg,
                calc_cfg,
                opt_cfg,
                lbfgs_cfg,
                rfo_cfg,
                bias_cfg,
                kind,
                freeze,
                out_dir_path,
            ) = _build_scan_context(
                yaml_cfg=yaml_cfg,
                geom_kw=dict(GEOM_KW_DEFAULT),
                calc_kw=dict(UMA_CALC_KW),
                opt_kw=dict(OPT_BASE_KW),
                lbfgs_kw=dict(LBFGS_KW),
                rfo_kw=dict(RFO_KW),
                bias_kw=dict(BIAS_KW),
                charge=resolved_charge,
                spin=resolved_spin,
                workers=workers,
                workers_per_node=workers_per_node,
                out_dir=out_dir,
                thresh=thresh,
                bias_k=float(bias_k),
                opt_mode=opt_mode,
                relax_max_cycles=relax_max_cycles,
                relax_override_requested=relax_override_requested,
                max_step_size=max_step_size,
                source_path=source_path,
                freeze_links=freeze_links,
            )

            if cli_param_overridden(ctx, "backend"):
                calc_cfg["backend"] = backend
            if cli_param_overridden(ctx, "solvent"):
                calc_cfg["solvent"] = solvent
            if cli_param_overridden(ctx, "solvent_model"):
                calc_cfg["solvent_model"] = solvent_model

            pdb_atom_meta: List[Dict[str, Any]] = []
            if source_path.suffix.lower() == ".pdb":
                pdb_atom_meta = load_pdb_atom_metadata(source_path)

            if scan_list_raw is None:
                raise click.BadParameter("--scan-lists is required.")
            scan_one_based = bool(one_based)
            scan_source = "--scan-lists"
            if is_scan_spec_file(scan_list_raw):
                spec_path = Path(scan_list_raw)
                parsed, raw_pairs, scan_one_based = parse_scan_spec_quads(
                    spec_path,
                    expected_len=2,
                    one_based_default=one_based,
                    atom_meta=pdb_atom_meta,
                    option_name="--scan-lists",
                )
                scan_source = f"--scan-lists ({spec_path})"
            else:
                parsed, raw_pairs = parse_scan_list_quads_checked(
                    scan_list_raw,
                    expected_len=2,
                    one_based=scan_one_based,
                    atom_meta=pdb_atom_meta,
                    option_name="--scan-lists",
                )
            (i1, j1, low1, high1), (i2, j2, low2, high2) = parsed
            d1_label_csv = axis_label_csv("d1", i1, j1, scan_one_based, pdb_atom_meta, raw_pairs[0])
            d2_label_csv = axis_label_csv("d2", i2, j2, scan_one_based, pdb_atom_meta, raw_pairs[1])
            d1_label_html = axis_label_html(d1_label_csv)
            d2_label_html = axis_label_html(d2_label_csv)
            if print_parsed:
                click.echo(
                    pretty_block(
                        "scan-parsed",
                        {
                            "source": scan_source,
                            "one_based": bool(scan_one_based),
                            "pairs_0based": parsed,
                        },
                    )
                )
            click.echo(
                pretty_block(
                    "scan-list (0-based)",
                    {"d1": (i1, j1, low1, high1), "d2": (i2, j2, low2, high2)},
                )
            )

            if pdb_atom_meta:
                click.echo("[scan2d] PDB atom details for scanned pairs:")
                legend = format_pdb_atom_metadata_header()
                click.echo(f"        legend: {legend}")
                click.echo(f"  d1 i: {format_pdb_atom_metadata(pdb_atom_meta, i1)}")
                click.echo(f"     j: {format_pdb_atom_metadata(pdb_atom_meta, j1)}")
                click.echo(f"  d2 i: {format_pdb_atom_metadata(pdb_atom_meta, i2)}")
                click.echo(f"     j: {format_pdb_atom_metadata(pdb_atom_meta, j2)}")

            # Temporary and grid directories
            tmp_root = Path(tempfile.mkdtemp(prefix="scan2d_tmp_"))
            grid_dir = out_dir_path / "grid"
            tmp_opt_dir = tmp_root / "opt"
            ensure_dir(grid_dir)
            ensure_dir(tmp_opt_dir)

            final_dir = out_dir_path

            coord_type = geom_cfg.get("coord_type", GEOM_KW_DEFAULT["coord_type"])
            geom_outer = geom_loader(
                geom_input_path, coord_type=coord_type, freeze_atoms=freeze
            )
            set_freeze_atoms_or_warn(geom_outer, freeze, context="scan2d")

            base_calc = create_calculator(**calc_cfg)
            biased = HarmonicBiasCalculator(base_calc, k=float(bias_cfg["k"]))

            # Records (including preopt) will be accumulated here
            records: List[Dict[str, Any]] = []
            ref_pdb_path = source_path if source_path.suffix.lower() == ".pdb" else None

            # Reference distances from the (pre)optimized structure, used for scan ordering
            d1_ref: Optional[float] = None
            d2_ref: Optional[float] = None

            # Cache of previously converged geometries for nearest-start logic:
            # each entry is (d1_A, d2_A, geometry_snapshot)
            visited_geoms: List[Tuple[float, float, Any]] = []

            if preopt:
                click.echo("[preopt] Unbiased relaxation of the initial structure ...")
                geom_outer.set_calculator(base_calc)
                max_step_bohr_local = float(max_step_size) * ANG2BOHR
                optimizer0 = make_sopt_optimizer(
                    geom_outer,
                    kind,
                    lbfgs_cfg,
                    rfo_cfg,
                    opt_cfg,
                    max_step_bohr=max_step_bohr_local,
                    relax_max_cycles=relax_max_cycles,
                    relax_override_requested=relax_override_requested,
                    out_dir=tmp_opt_dir,
                    prefix="preopt",
                )
                try:
                    optimizer0.run()
                except ZeroStepLength:
                    click.echo("[preopt] ZeroStepLength — continuing.")
                except OptimizationError as e:
                    click.echo(f"[preopt] OptimizationError — {e}")

                # Measure optimized distances and record preopt structure
                try:
                    coords_outer = np.asarray(getattr(geom_outer, "coords"), dtype=float).reshape(-1, 3)
                    d1_ref = distance_A_from_coords(coords_outer, i1, j1)
                    d2_ref = distance_A_from_coords(coords_outer, i2, j2)

                    d1_tag = distance_tag(d1_ref)
                    d2_tag = distance_tag(d2_ref)

                    preopt_xyz_path = grid_dir / f"preopt_i{d1_tag}_j{d2_tag}.xyz"
                    s = geom_outer.as_xyz()
                    if not s.endswith("\n"):
                        s += "\n"
                    with open(preopt_xyz_path, "w") as f:
                        f.write(s)

                    convert_xyz_like_outputs(
                        preopt_xyz_path,
                        prepared_input,
                        ref_pdb_path=ref_pdb_path,
                        out_pdb_path=grid_dir / f"preopt_i{d1_tag}_j{d2_tag}.pdb",
                        out_gjf_path=grid_dir / f"preopt_i{d1_tag}_j{d2_tag}.gjf",
                        context=f"'{preopt_xyz_path.name}' to PDB/GJF",
                    )

                    E_pre_h = unbiased_energy_hartree(geom_outer, base_calc)
                    records.append(
                        {
                            "i": int(-1),
                            "j": int(-1),
                            "d1_A": float(d1_ref),
                            "d2_A": float(d2_ref),
                            "energy_hartree": E_pre_h,
                            "bias_converged": True,
                        }
                    )
                    # Store preoptimized geometry as a candidate for nearest-start
                    visited_geoms.append(
                        (float(d1_ref), float(d2_ref), _snapshot_geometry(geom_outer))
                    )

                    click.echo(
                        f"[preopt] Recorded preoptimized structure at d1={d1_ref:.3f} Å, d2={d2_ref:.3f} Å."
                    )
                except Exception as e:
                    click.echo(
                        f"[preopt] WARNING: failed to record preoptimized structure: {e}",
                        err=True,
                    )

            max_step_bohr = float(max_step_size) * ANG2BOHR

            # Construct scan grids and reorder so that points near the preopt geometry are visited first
            d1_values = values_from_bounds(low1, high1, float(max_step_size))
            d2_values = values_from_bounds(low2, high2, float(max_step_size))

            d1_values = _sort_values_by_reference(d1_values, d1_ref)
            d2_values = _sort_values_by_reference(d2_values, d2_ref)

            N1, N2 = len(d1_values), len(d2_values)
            click.echo(
                f"[grid] d1 steps = {N1}  values(A)={list(map(lambda x:f'{x:.3f}', d1_values))}"
            )
            click.echo(
                f"[grid] d2 steps = {N2}  values(A)={list(map(lambda x:f'{x:.3f}', d2_values))}"
            )
            click.echo(f"[grid] total grid points = {N1*N2}")

            # Track the start index of the previous row's entries in visited_geoms
            # so we can prune old rows and bound memory usage.
            _prev_row_start: int = 0

            for i_idx, d1_target in enumerate(d1_values):
                # Prune visited_geoms: keep only the previous row's entries
                # (current row hasn't started yet).  Entries before _prev_row_start
                # belong to even older rows and are unlikely nearest neighbors.
                if _prev_row_start > 0:
                    visited_geoms = visited_geoms[_prev_row_start:]
                    _prev_row_start = 0
                # Mark where the current row starts; this becomes _prev_row_start
                # at the beginning of the *next* outer iteration.
                _cur_row_start = len(visited_geoms)

                click.echo(
                    f"[stage] d1 step {i_idx+1}/{N1}: target = {d1_target:.3f} Å"
                )
                biased.set_pairs([(i1, j1, float(d1_target))])
                geom_outer.set_calculator(biased)

                opt1 = make_sopt_optimizer(
                    geom_outer,
                    kind,
                    lbfgs_cfg,
                    rfo_cfg,
                    opt_cfg,
                    max_step_bohr=max_step_bohr,
                    relax_max_cycles=relax_max_cycles,
                    relax_override_requested=relax_override_requested,
                    out_dir=tmp_opt_dir,
                    prefix=f"d1_{i_idx:03d}",
                )
                try:
                    opt1.run()
                except ZeroStepLength:
                    click.echo(
                        f"[d1 {i_idx}] ZeroStepLength — continuing to d2 scan."
                    )
                except OptimizationError as e:
                    click.echo(f"[d1 {i_idx}] OptimizationError — {e}")

                geom_inner = _snapshot_geometry(geom_outer)
                geom_inner.set_calculator(biased)

                # Store the d1-relaxed structure as a candidate for nearest-start
                try:
                    coords_inner = np.asarray(getattr(geom_inner, "coords"), dtype=float).reshape(-1, 3)
                    d1_cur = distance_A_from_coords(coords_inner, i1, j1)
                    d2_cur = distance_A_from_coords(coords_inner, i2, j2)
                    visited_geoms.append(
                        (float(d1_cur), float(d2_cur), _snapshot_geometry(geom_inner))
                    )
                except Exception as e:
                    click.echo(
                        f"[nearest-start] WARNING: failed to store d1-relaxed structure for d1={d1_target:.3f} Å: {e}",
                        err=True,
                    )

                trj_blocks = [] if dump else None

                for j_idx, d2_target in enumerate(d2_values):
                    # Choose initial structure: nearest previously converged (d1,d2) point
                    if visited_geoms:
                        try:
                            target_vec = np.array(
                                [float(d1_target), float(d2_target)], dtype=float
                            )
                            prev_coords = np.array(
                                [(g[0], g[1]) for g in visited_geoms],
                                dtype=float,
                            )
                            dists2 = np.sum((prev_coords - target_vec) ** 2, axis=1)
                            best_idx = int(np.argmin(dists2))
                            _, _, best_geom = visited_geoms[best_idx]
                            # Reset geom_inner coordinates to the best previous geometry
                            try:
                                geom_inner.coords[:] = np.array(
                                    best_geom.coords, copy=True
                                )
                            except Exception:
                                geom_inner.coords = np.array(
                                    best_geom.coords, copy=True
                                )
                        except Exception as e:
                            click.echo(
                                f"[nearest-start] WARNING: failed to select nearest previous structure for d1={d1_target:.3f}, d2={d2_target:.3f}: {e}",
                                err=True,
                            )

                    biased.set_pairs(
                        [
                            (i1, j1, float(d1_target)),
                            (i2, j2, float(d2_target)),
                        ]
                    )
                    geom_inner.set_calculator(biased)

                    opt2 = make_sopt_optimizer(
                        geom_inner,
                        kind,
                        lbfgs_cfg,
                        rfo_cfg,
                        opt_cfg,
                        max_step_bohr=max_step_bohr,
                        relax_max_cycles=relax_max_cycles,
                        relax_override_requested=relax_override_requested,
                        out_dir=tmp_opt_dir,
                        prefix=f"d1_{i_idx:03d}_d2_{j_idx:03d}",
                    )
                    try:
                        opt2.run()
                        converged = True
                    except ZeroStepLength:
                        click.echo(
                            f"[d1 {i_idx}, d2 {j_idx}] ZeroStepLength — recorded anyway.",
                            err=True,
                        )
                        converged = False
                    except OptimizationError as e:
                        click.echo(
                            f"[d1 {i_idx}, d2 {j_idx}] OptimizationError — {e}"
                        )
                        converged = False

                    E_h = unbiased_energy_hartree(geom_inner, base_calc)

                    # Write per-grid XYZ snapshots under result_scan2d/grid/
                    d1_tag = distance_tag(d1_target)
                    d2_tag = distance_tag(d2_target)
                    xyz_path = grid_dir / f"point_i{d1_tag}_j{d2_tag}.xyz"
                    try:
                        s = geom_inner.as_xyz()
                        if not s.endswith("\n"):
                            s += "\n"
                        with open(xyz_path, "w") as f:
                            f.write(s)
                        convert_xyz_like_outputs(
                            xyz_path,
                            prepared_input,
                            ref_pdb_path=ref_pdb_path,
                            out_pdb_path=grid_dir / f"point_i{d1_tag}_j{d2_tag}.pdb",
                            out_gjf_path=grid_dir / f"point_i{d1_tag}_j{d2_tag}.gjf",
                            context=f"'{xyz_path.name}' to PDB/GJF",
                        )
                    except Exception as e:
                        click.echo(
                            f"[write] WARNING: failed to write {xyz_path.name}: {e}",
                            err=True,
                        )

                    # Store this converged grid point for nearest-start initialization
                    try:
                        coords_inner = np.asarray(getattr(geom_inner, "coords"), dtype=float).reshape(-1, 3)
                        d1_cur = distance_A_from_coords(coords_inner, i1, j1)
                        d2_cur = distance_A_from_coords(coords_inner, i2, j2)
                        visited_geoms.append(
                            (float(d1_cur), float(d2_cur), _snapshot_geometry(geom_inner))
                        )
                    except Exception as e:
                        click.echo(
                            f"[nearest-start] WARNING: failed to store geometry for d1={d1_target:.3f}, d2={d2_target:.3f}: {e}",
                            err=True,
                        )

                    if dump and trj_blocks is not None:
                        sblock = geom_inner.as_xyz()
                        if not sblock.endswith("\n"):
                            sblock += "\n"
                        trj_blocks.append(sblock)

                    records.append(
                        {
                            "i": int(i_idx),
                            "j": int(j_idx),
                            "d1_A": float(d1_target),
                            "d2_A": float(d2_target),
                            "energy_hartree": E_h,
                            "bias_converged": bool(converged),
                        }
                    )

                # Update row tracking: current row entries start at _cur_row_start
                _prev_row_start = _cur_row_start

                if dump and trj_blocks:
                    trj_path = grid_dir / f"inner_path_d1_{i_idx:03d}_trj.xyz"
                    try:
                        with open(trj_path, "w") as f:
                            f.write("".join(trj_blocks))
                        click.echo(f"[write] Wrote '{trj_path}'.")
                        convert_xyz_like_outputs(
                            trj_path,
                            prepared_input,
                            ref_pdb_path=ref_pdb_path,
                            out_pdb_path=grid_dir / f"inner_path_d1_{i_idx:03d}.pdb",
                            context=f"'{trj_path.name}' to PDB",
                        )
                    except Exception as e:
                        click.echo(
                            f"[write] WARNING: failed to write '{trj_path}': {e}"
                        )

            # ===== surface.csv (final output directly under result_scan2d) =====
            df = pd.DataFrame.from_records(records)
            if df.empty:
                click.echo("No grid records produced; aborting.")
                sys.exit(1)

            if baseline == "first":
                ref = float(
                    df.loc[(df["i"] == 0) & (df["j"] == 0), "energy_hartree"].iloc[0]
                )
            else:
                ref = float(df["energy_hartree"].min())
            df["energy_kcal"] = (df["energy_hartree"] - ref) * AU2KCALPERMOL
            df["d1_label"] = d1_label_csv
            df["d2_label"] = d2_label_csv

            surface_csv = final_dir / "surface.csv"
            df.to_csv(surface_csv, index=False)
            click.echo(f"[write] Wrote '{surface_csv}'.")

            # ===== Plots (RBF on a fixed 50×50 grid, unified layout, placed under final_dir) =====
            d1_points = df["d1_A"].to_numpy(dtype=float)
            d2_points = df["d2_A"].to_numpy(dtype=float)
            z_points = df["energy_kcal"].to_numpy(dtype=float)
            mask = (
                np.isfinite(d1_points)
                & np.isfinite(d2_points)
                & np.isfinite(z_points)
            )
            if not np.any(mask):
                click.echo("[plot] No finite data for plotting.")
                sys.exit(1)

            x_min, x_max = float(np.min(d1_points[mask])), float(
                np.max(d1_points[mask])
            )
            y_min, y_max = float(np.min(d2_points[mask])), float(
                np.max(d2_points[mask])
            )

            xi = np.linspace(x_min, x_max, 50)
            yi = np.linspace(y_min, y_max, 50)
            XI, YI = np.meshgrid(xi, yi)

            rbf = Rbf(
                d1_points[mask], d2_points[mask], z_points[mask], function="multiquadric"
            )
            ZI = rbf(XI, YI)

            vmin = float(np.nanmin(ZI)) if zmin is None else float(zmin)
            vmax = float(np.nanmax(ZI)) if zmax is None else float(zmax)
            if (
                not np.isfinite(vmin)
                or not np.isfinite(vmax)
                or vmax <= vmin
            ):
                vmin, vmax = float(np.nanmin(ZI)), float(np.nanmax(ZI))

            # Choose neat contour/tick steps
            def _nice_step(span: float) -> float:
                if span <= 0:
                    return 1.0
                raw = span / 6.0
                mag = 10 ** math.floor(math.log10(raw))
                candidates = (0.5, 1, 2, 5, 10, 20)
                best = candidates[0] * mag
                best_err = abs(best - raw)
                for m in candidates[1:]:
                    s = m * mag
                    err = abs(s - raw)
                    if err < best_err:
                        best, best_err = s, err
                return best

            c_step = _nice_step(vmax - vmin)
            c_start = math.floor(vmin / c_step) * c_step
            c_end = math.ceil(vmax / c_step) * c_step

            # ---- 2D contour plot (PNG with explicit size) ----
            fig2d = go.Figure(
                data=go.Contour(
                    z=ZI,
                    x=xi,
                    y=yi,
                    contours=dict(start=c_start, end=c_end, size=c_step),
                    zmin=vmin,
                    zmax=vmax,
                    contours_coloring="heatmap",
                    colorscale="plasma",
                    colorbar=dict(
                        title=dict(
                            text="(kcal/mol)", side="top", font=dict(size=16, color="#1C1C1C")
                        ),
                        tickfont=dict(size=14, color="#1C1C1C"),
                        ticks="inside",
                        ticklen=10,
                        tickcolor="#1C1C1C",
                        outlinecolor="#1C1C1C",
                        outlinewidth=2,
                        lenmode="fraction",
                        len=1.11,
                        x=1.05,
                        y=0.53,
                        xanchor="left",
                        yanchor="middle",
                    ),
                )
            )
            fig2d.update_layout(
                width=640,
                height=600,
                xaxis_title=d1_label_html,
                yaxis_title=d2_label_html,
                plot_bgcolor="white",
                xaxis=dict(
                    range=[x_min, x_max],
                    showline=True,
                    linewidth=3,
                    linecolor="#1C1C1C",
                    mirror=True,
                    tickson="boundaries",
                    ticks="inside",
                    tickwidth=3,
                    tickcolor="#1C1C1C",
                    title_font=dict(size=18, color="#1C1C1C"),
                    tickfont=dict(size=18, color="#1C1C1C"),
                    tickvals=list(np.linspace(x_min, x_max, 6)),
                    tickformat=".2f",
                ),
                yaxis=dict(
                    range=[y_min, y_max],
                    showline=True,
                    linewidth=3,
                    linecolor="#1C1C1C",
                    mirror=True,
                    tickson="boundaries",
                    ticks="inside",
                    tickwidth=3,
                    tickcolor="#1C1C1C",
                    title_font=dict(size=18, color="#1C1C1C"),
                    tickfont=dict(size=18, color="#1C1C1C"),
                    tickvals=list(np.linspace(y_min, y_max, 6)),
                    tickformat=".2f",
                ),
                margin=dict(l=10, r=10, b=10, t=40),
            )
            png2d = final_dir / "scan2d_map.png"
            fig2d.write_image(str(png2d), scale=2, engine="kaleido", width=680, height=600)
            click.echo(f"[plot] Wrote '{png2d}'.")

            # ---- 3D surface plus base-plane projection ----
            spread = vmax - vmin if (vmax > vmin) else 1.0
            z_bottom = vmin - spread
            z_top = vmax

            # Avoid ticks below zmin (= vmin) and snap to sensible values
            z_step = _nice_step(vmax - vmin)
            z_start_tick = math.ceil(vmin / z_step) * z_step  # First tick must be ≥ vmin
            z_ticks = np.arange(z_start_tick, z_top + 0.5 * z_step, z_step).tolist()

            surface3d = go.Surface(
                x=XI,
                y=YI,
                z=ZI,
                colorscale="plasma",
                cmin=vmin,
                cmax=vmax,
                colorbar=dict(
                    title=dict(
                        text="(kcal/mol)", side="top", font=dict(size=16, color="#1C1C1C")
                    ),
                    tickfont=dict(size=14, color="#1C1C1C"),
                    ticks="inside",
                    ticklen=10,
                    tickcolor="#1C1C1C",
                    outlinecolor="#1C1C1C",
                    outlinewidth=2,
                    lenmode="fraction",
                    len=1.11,
                    x=1.05,
                    y=0.53,
                    xanchor="left",
                    yanchor="middle",
                ),
                contours={
                    "z": {
                        "show": True,
                        "start": c_start,
                        "end": c_end,
                        "size": c_step,
                        "color": "black",
                        "project": {"z": True},
                    }
                },
                name="3D Surface",
            )

            plane_proj = go.Surface(
                x=XI,
                y=YI,
                z=np.full_like(ZI, z_bottom),
                surfacecolor=ZI,
                colorscale="plasma",
                cmin=vmin,
                cmax=vmax,
                showscale=False,
                opacity=1.0,
                name="2D Contour Projection (Bottom)",
            )

            fig3d = go.Figure(data=[surface3d, plane_proj])
            fig3d.update_layout(
                title="Energy Landscape with 2D PES Scan",
                width=800,
                height=700,
                scene=dict(
                    bgcolor="rgba(0,0,0,0)",
                    xaxis=dict(
                        title=d1_label_html,
                        range=[x_min, x_max],
                        showline=True,
                        linewidth=4,
                        linecolor="#1C1C1C",
                        mirror=True,
                        ticks="inside",
                        tickwidth=4,
                        tickcolor="#1C1C1C",
                        gridcolor="rgba(0,0,0,0.1)",
                        zerolinecolor="rgba(0,0,0,0.1)",
                        showbackground=False,
                    ),
                    yaxis=dict(
                        title=d2_label_html,
                        range=[y_min, y_max],
                        showline=True,
                        linewidth=4,
                        linecolor="#1C1C1C",
                        mirror=True,
                        ticks="inside",
                        tickwidth=4,
                        tickcolor="#1C1C1C",
                        gridcolor="rgba(0,0,0,0.1)",
                        zerolinecolor="rgba(0,0,0,0.1)",
                        showbackground=False,
                    ),
                    zaxis=dict(
                        title="Potential Energy (kcal/mol)",
                        range=[z_bottom, z_top],
                        tickmode="array",
                        tickvals=z_ticks,
                        showline=True,
                        linewidth=4,
                        linecolor="#1C1C1C",
                        mirror=True,
                        ticks="inside",
                        tickwidth=4,
                        tickcolor="#1C1C1C",
                        showgrid=True,
                        gridcolor="rgba(0,0,0,0.1)",
                        zerolinecolor="rgba(0,0,0,0.1)",
                        showbackground=False,
                    ),
                ),
                margin=dict(l=10, r=20, b=10, t=40),
                paper_bgcolor="white",
            )

            html3d = final_dir / "scan2d_landscape.html"
            fig3d.write_html(str(html3d))
            click.echo(f"[plot] Wrote '{html3d}'.")

            click.echo("====== 2D Scan finished ======\n")
            click.echo(format_elapsed("[time] Elapsed Time for 2D Scan", time_start))

        except KeyboardInterrupt:
            click.echo("Interrupted by user.", err=True)
            sys.exit(130)
        except Exception as e:
            tb = "".join(traceback.format_exception(type(e), e, e.__traceback__))
            click.echo(
                "Unhandled exception during 2D scan:\n"
                + textwrap.indent(tb, "  "),
                err=True,
            )
            sys.exit(1)
        finally:
            if tmp_root is not None:
                shutil.rmtree(tmp_root, ignore_errors=True)



# tests/test_defaults.py
"""Tests for pdb2reaction.defaults configuration constants."""

import pytest

from pdb2reaction.defaults import (
    GEOM_KW_DEFAULT,
    CALC_KW_DEFAULT,
    OPT_BASE_KW,
    LBFGS_KW,
    RFO_KW,
    IRC_KW,
    BOND_KW,
    BIAS_KW,
    THERMO_KW,
    OPT_MODE_ALIASES,
    TSOPT_MODE_ALIASES,
)


class TestDefaultsStructure:
    """Verify that default dictionaries have expected keys and types."""

    def test_geom_kw_has_coord_type(self):
        assert "coord_type" in GEOM_KW_DEFAULT
        assert isinstance(GEOM_KW_DEFAULT["coord_type"], str)

    def test_calc_kw_has_backend(self):
        assert "backend" in CALC_KW_DEFAULT
        assert CALC_KW_DEFAULT["backend"] == "uma"

    def test_calc_kw_has_charge_spin(self):
        assert "charge" in CALC_KW_DEFAULT
        assert "spin" in CALC_KW_DEFAULT

    def test_opt_base_kw_has_thresh(self):
        assert "thresh" in OPT_BASE_KW
        assert isinstance(OPT_BASE_KW["thresh"], str)

    def test_lbfgs_has_max_cycles(self):
        assert "max_cycles" in LBFGS_KW
        assert LBFGS_KW["max_cycles"] > 0

    def test_rfo_has_max_cycles(self):
        assert "max_cycles" in RFO_KW
        assert RFO_KW["max_cycles"] > 0

    def test_irc_kw_has_step_length(self):
        assert "step_length" in IRC_KW
        assert IRC_KW["step_length"] > 0

    def test_bond_kw_has_factor(self):
        assert "bond_factor" in BOND_KW
        assert BOND_KW["bond_factor"] > 0

    def test_bias_kw_has_k(self):
        assert "k" in BIAS_KW
        assert float(BIAS_KW["k"]) > 0

    def test_thermo_kw_structure(self):
        assert "temperature" in THERMO_KW
        assert THERMO_KW["temperature"] > 0

    def test_opt_mode_aliases_cover_expected(self):
        # OPT_MODE_ALIASES is a tuple of ((aliases,...), canonical) pairs
        canonical_modes = {pair[1] for pair in OPT_MODE_ALIASES}
        assert "lbfgs" in canonical_modes
        assert "rfo" in canonical_modes

    def test_tsopt_mode_aliases_cover_expected(self):
        canonical_modes = {pair[1] for pair in TSOPT_MODE_ALIASES}
        assert "dimer" in canonical_modes
        assert "rsirfo" in canonical_modes


class TestDefaultsConsistency:
    """Cross-check defaults for internal consistency."""

    def test_no_negative_cycles(self):
        for name, d in [("LBFGS_KW", LBFGS_KW), ("RFO_KW", RFO_KW), ("OPT_BASE_KW", OPT_BASE_KW)]:
            if "max_cycles" in d:
                assert d["max_cycles"] > 0, f"{name} has non-positive max_cycles"

    def test_solvent_default_is_none(self):
        assert CALC_KW_DEFAULT.get("solvent", "none") == "none"



# tests/test_cli_utils.py
"""Tests for pdb2reaction.cli_utils."""

import os
import shutil
import tempfile
from pathlib import Path

import click
import pytest
import yaml

from pdb2reaction.cli_utils import (
    resolve_yaml_sources,
    load_merged_yaml_cfg,
    link_or_copy_file,
)


class TestResolveYamlSources:
    def test_no_sources(self):
        config, override, legacy = resolve_yaml_sources(None, None, None)
        assert config is None
        assert override is None
        assert legacy is False

    def test_config_only(self):
        p = Path("config.yaml")
        config, override, legacy = resolve_yaml_sources(p, None, None)
        assert config is p
        assert override is None
        assert legacy is False

    def test_override_only(self):
        p = Path("override.yaml")
        config, override, legacy = resolve_yaml_sources(None, p, None)
        assert config is None
        assert override is p
        assert legacy is False

    def test_legacy_only(self):
        p = Path("legacy.yaml")
        config, override, legacy = resolve_yaml_sources(None, None, p)
        assert config is None
        assert override is p
        assert legacy is True

    def test_config_and_override(self):
        c = Path("config.yaml")
        o = Path("override.yaml")
        config, override, legacy = resolve_yaml_sources(c, o, None)
        assert config is c
        assert override is o
        assert legacy is False

    def test_conflict_raises(self):
        with pytest.raises(click.BadParameter):
            resolve_yaml_sources(None, Path("a.yaml"), Path("b.yaml"))


class TestLoadMergedYamlCfg:
    def test_no_files(self):
        merged, cfg, ovr = load_merged_yaml_cfg(None, None)
        assert merged == {}
        assert cfg == {}
        assert ovr == {}

    def test_single_config(self, tmp_path):
        cfg_file = tmp_path / "config.yaml"
        cfg_file.write_text(yaml.dump({"opt": {"max_cycles": 100}}))
        merged, cfg, ovr = load_merged_yaml_cfg(cfg_file, None)
        assert merged["opt"]["max_cycles"] == 100
        assert ovr == {}

    def test_merge_override(self, tmp_path):
        cfg_file = tmp_path / "config.yaml"
        cfg_file.write_text(yaml.dump({"opt": {"max_cycles": 100, "thresh": "gau"}}))
        ovr_file = tmp_path / "override.yaml"
        ovr_file.write_text(yaml.dump({"opt": {"max_cycles": 50}}))
        merged, cfg, ovr = load_merged_yaml_cfg(cfg_file, ovr_file)
        assert merged["opt"]["max_cycles"] == 50  # Override wins
        assert merged["opt"]["thresh"] == "gau"  # Config preserved


class TestLinkOrCopyFile:
    def test_creates_symlink(self, tmp_path):
        src = tmp_path / "src.txt"
        src.write_text("hello")
        dst = tmp_path / "dst.txt"
        result = link_or_copy_file(src, dst)
        assert result is True
        assert dst.exists()
        assert dst.read_text() == "hello"

    def test_overwrites_existing(self, tmp_path):
        src = tmp_path / "src.txt"
        src.write_text("new")
        dst = tmp_path / "dst.txt"
        dst.write_text("old")
        result = link_or_copy_file(src, dst)
        assert result is True
        assert dst.read_text() == "new"

    def test_directory_destination_returns_false(self, tmp_path):
        src = tmp_path / "src.txt"
        src.write_text("hello")
        dst = tmp_path / "dst_dir"
        dst.mkdir()
        result = link_or_copy_file(src, dst)
        assert result is False



import logging

import numpy as np

from pysisyphus.Geometry import Geometry
from pysisyphus.helpers_pure import log
from pysisyphus.intcoords.setup import get_bond_sets
from pysisyphus.intcoords import RedundantCoords
from pysisyphus.intcoords.PrimTypes import PrimTypes


logger = logging.getLogger("internal_coords")


def augment_bonds(geom, root=0, proj=False):
    assert geom.coord_type != "cart"
    log(logger, "Trying to augment bonds.")

    hessian = geom.cart_hessian
    try:
        energy = geom.energy
    except AttributeError:
        energy = None

    func = find_missing_bonds_by_projection if proj else find_missing_strong_bonds

    missing_bonds = func(geom, hessian, root=root)

    if missing_bonds:
        aux_bond_pt = PrimTypes.AUX_BOND
        missing_aux_bonds = [(aux_bond_pt, *mbond) for mbond in missing_bonds]
        print("\t@Missing bonds:", missing_bonds)
        new_geom = Geometry(geom.atoms, geom.cart_coords,
                            coord_type=geom.coord_type,
                            coord_kwargs={"define_prims": missing_aux_bonds,},
        )
        new_geom.set_calculator(geom.calculator)
        new_geom.energy = energy
        new_geom.cart_hessian = hessian
        return new_geom
    else:
        return geom


def find_missing_strong_bonds(geom, hessian, bond_factor=1.7, thresh=0.3,
                              root=0):
    # Define only bonds
    red = RedundantCoords(geom.atoms, geom.cart_coords,
                          bond_factor=bond_factor, bonds_only=True)
    cur_bonds = set([frozenset(b) for b in geom.internal.bond_atom_indices])

    # Transform cartesian hessian to bond hessian
    bond_hess = red.transform_hessian(hessian)
    # Determine transisiton vector
    eigvals, eigvecs = np.linalg.eigh(bond_hess)
    # There are probably no bonds missing if there are no negative eigenvalues
    if sum(eigvals < 0) == 0:
        return list()

    trans_vec = eigvecs[:, root]
    # Find bonds that strongly contribute to the selected transition vector
    strong = np.abs(trans_vec) > thresh
    strong_bonds = np.array(red.bond_atom_indices)[strong]
    strong_bonds = set([frozenset(b) for b in strong_bonds])

    # Check which strong bonds are missing from the currently defiend bonds
    missing_bonds = strong_bonds - cur_bonds
    missing_bonds = [tuple(_) for _ in missing_bonds]
    return missing_bonds


def find_missing_bonds_by_projection(geom, hessian, bond_factor=2.0, bond_thresh=0.35,
                                     concerted_thresh=0.35, root=0):

    def array2set(arr):
        return set([tuple(_) for _ in arr])

    bonds_present = array2set(geom.internal.bond_atom_indices)
    eigvals, eigvecs = np.linalg.eigh(hessian)

    # There are probably no bonds missing if there are no negative eigenvalues
    if sum(eigvals < 0) == 0:
        return list()

    trans_vec = eigvecs[:, root]

    c3d = geom.coords3d
    bond_vec_empty = np.zeros_like(c3d)
    unique_bonds = array2set(get_bond_sets(geom.atoms, c3d, bond_factor=bond_factor))
    unique_bonds -= bonds_present
    unique_bonds = np.array(list(unique_bonds))

    bond_vecs = list()
    concerted_vecs = list()
    for m, k in unique_bonds:
        displ = c3d[k] - c3d[m]
        displ /= np.linalg.norm(displ)

        # Bond
        bond = bond_vec_empty.copy()
        bond[k] = displ
        bond[m] = -displ
        bond_vecs.append(bond)

        # Concerted movement
        conc = bond_vec_empty.copy()
        conc[k] = displ
        conc[m] = displ
        concerted_vecs.append(conc)

    def reshape(arr):
        return np.array(arr).reshape(-1, trans_vec.size)
    bond_vecs = reshape(bond_vecs)
    concerted_vecs = reshape(concerted_vecs)

    def overlaps(arr):
        return np.abs(arr.dot(trans_vec))
    bond_ovlps = overlaps(bond_vecs)
    concerted_ovlps = overlaps(concerted_vecs)

    unique_bonds = np.array(unique_bonds)
    missing_bonds = unique_bonds[bond_ovlps > bond_thresh]
    missing_concerted = unique_bonds[concerted_ovlps > concerted_thresh]

    missing_inds = array2set(missing_bonds) | array2set(missing_concerted)
    return missing_inds
